# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import Dict, List, Optional, Tuple, Set, cast

import networkx as nx  # type: ignore
import warnings

from discopop_explorer.PEGraphX import (
    PEGraphX,
    FunctionNode,
    EdgeType,
    LoopNode,
    CUNode,
    NodeID,
    MemoryRegion,
)
from discopop_explorer.utils import calculate_workload
from discopop_library.discopop_optimizer.PETParser.DataAccesses.FromCUs import (
    get_data_accesses_for_cu,
)
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.ContextMerge import ContextMerge
from discopop_library.discopop_optimizer.classes.nodes.ContextRestore import ContextRestore
from discopop_library.discopop_optimizer.classes.nodes.ContextSave import ContextSave
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshot import ContextSnapshot
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshotPop import ContextSnapshotPop
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    add_dataflow_edge,
    get_all_nodes_in_function,
    get_all_parents,
    get_nodes_by_functions,
    get_parents,
    get_path_entry,
    get_predecessors,
    get_successors,
    get_children,
    add_successor_edge,
    add_child_edge,
    add_temporary_edge,
    redirect_edge,
    convert_temporary_edges,
    get_all_function_nodes,
    get_read_and_written_data_from_subgraph,
    show,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


class PETParser(object):
    pet: PEGraphX
    graph: nx.DiGraph
    next_free_node_id: int
    cu_id_to_graph_node_id: Dict[NodeID, int]
    experiment: Experiment
    in_data_flow: Dict[int, Set[int]]
    out_data_flow: Dict[int, Set[int]]

    def __init__(self, experiment: Experiment):
        self.pet = experiment.detection_result.pet
        self.graph = nx.DiGraph()
        self.next_free_node_id = 0
        self.cu_id_to_graph_node_id = dict()
        self.experiment = experiment

    def parse(self) -> Tuple[nx.DiGraph, int]:
        if self.experiment.arguments.verbose:
            print("PARSING PET")
        self.__add_cu_nodes()
        if self.experiment.arguments.verbose:
            print("added cu nodes")
        self.__add_functions()
        if self.experiment.arguments.verbose:
            print("added functions")
        self.__add_pet_successor_edges()
        if self.experiment.arguments.verbose:
            print("added successor edges")
        self.__add_loop_nodes()
        if self.experiment.arguments.verbose:
            print("added loop nodes")
        # self.__add_branch_return_node()
        self.__add_function_return_node()

        self.__new_parse_branched_sections()

        convert_temporary_edges(self.graph)
        if self.experiment.arguments.verbose:
            print("converted temporary edges")

        #        self.__mark_branch_affiliation()
        #        print("marked branch affiliations")
        self.__calculate_data_flow()
        if self.experiment.arguments.verbose:
            print("calculated data flow")

        self.__propagate_reads_and_writes()
        if self.experiment.arguments.verbose:
            print("Propagated read/write information")

        return self.graph, self.next_free_node_id

    def get_new_node_id(self) -> int:
        """returns the next free node id and increments the counter"""
        buffer = self.next_free_node_id
        self.next_free_node_id += 1
        return buffer

    def __add_function_return_node(self):
        function_node_ids = get_all_function_nodes(self.graph)
        dummy_return_nodes: Set[int] = set()
        function_return_nodes: Dict[int, int] = dict()
        for function in function_node_ids:
            return_dummy_id = self.get_new_node_id()
            self.graph.add_node(return_dummy_id, data=Workload(return_dummy_id, None, None, None, None))
            function_return_nodes[function] = return_dummy_id
            dummy_return_nodes.add(return_dummy_id)

        for node in self.graph.nodes():
            if node in dummy_return_nodes:
                continue
            if len(get_successors(self.graph, node)) == 0:
                # node is end of path
                # check if node is contained in function
                parent_functions = [e for e in get_all_parents(self.graph, node) if e in function_node_ids]
                if len(parent_functions) > 0:
                    for parent_func in parent_functions:
                        # connect end of path to the dummy return node
                        add_successor_edge(self.graph, node, function_return_nodes[parent_func])
                        if self.experiment.arguments.verbose:
                            print("ADDED DUMMY CONNECTION: ", node, function_return_nodes[parent_func])

    def __add_branch_return_node(self):
        """makes sure every branching section has a merge node"""
        path_return_nodes: Dict[int, int] = dict()

        for node in copy.deepcopy(self.graph.nodes()):
            if len(get_successors(self.graph, node)) == 0:
                path_entry = get_path_entry(self.graph, node)
                if path_entry not in path_return_nodes:
                    # create new dummy return node
                    dummy_return_node_id = self.get_new_node_id()
                    self.graph.add_node(
                        dummy_return_node_id, data=Workload(dummy_return_node_id, self.experiment, None, None, None)
                    )
                    path_return_nodes[path_entry] = dummy_return_node_id
                # connect to existing return node
                add_successor_edge(self.graph, node, path_return_nodes[path_entry])
                print("ADDED EDGE: ", node, "->", path_return_nodes[path_entry])

    def __new_parse_branched_sections(self):
        """Branched sections in the CU Graph are represented by a serialized version in the MOG.
        To make this possible, Context Snapshot, Restore and Merge points are added to allow a synchronization
        'between' the different branches"""

        all_functions = get_all_function_nodes(self.graph)
        nodes_by_functions = get_nodes_by_functions(self.graph)
        for idx, function in enumerate(all_functions):
            if self.experiment.arguments.verbose:
                print("FUNCTION: ", data_at(self.graph, function).name, idx, "/", len(all_functions))
            nodes_in_function = nodes_by_functions[function]

            post_dominators = self.__get_post_dominators(nodes_in_function)

            path_splits = self.__get_path_splits(nodes_in_function)
            merge_nodes = self.__get_merge_nodes(path_splits, post_dominators)

            added_node_ids = self.__fix_empty_branches(merge_nodes, post_dominators)
            nodes_in_function = list(set(nodes_in_function).union(set(added_node_ids)))

            # re-calculate post_dominators and merge nodes
            #            post_dominators = self.__get_post_dominators(nodes_in_function)
            #            path_splits = self.__get_path_splits(nodes_in_function)
            #            merge_nodes = self.__get_merge_nodes(path_splits, post_dominators)

            self.__insert_context_nodes(nodes_in_function)

    def __fix_empty_branches(
        self, merge_nodes: Dict[int, Optional[int]], post_dominators: Dict[int, Set[int]]
    ) -> List[int]:
        """do not allow empty branches. Adds an empty dummy node if necessary"""
        empty_branches: Set[Tuple[int, int]] = set()
        for split_node in merge_nodes:
            if merge_nodes[split_node] is None:
                raise ValueError("No branching without merge allowed!")
            for branch_entry in get_successors(self.graph, split_node):
                for branch_exit in get_predecessors(self.graph, cast(int, merge_nodes[split_node])):
                    if branch_exit == split_node and branch_entry == merge_nodes[split_node]:
                        empty_branches.add((branch_exit, branch_entry))

        added_node_ids: List[int] = []
        for entry, exit in empty_branches:
            dummy_node_id = self.get_new_node_id()
            self.graph.add_node(dummy_node_id, data=Workload(dummy_node_id, self.experiment, None, None, None))
            if self.experiment.arguments.verbose:
                print("Added dummy node: ", entry, "->", dummy_node_id, "->", exit)
            redirect_edge(self.graph, entry, entry, exit, dummy_node_id)
            add_successor_edge(self.graph, dummy_node_id, exit)
            added_node_ids.append(dummy_node_id)
        return added_node_ids

    def __insert_context_nodes(self, node_list: List[int]):
        """flattens the graph via inserting context nodes"""
        modification_found = True
        while modification_found:
            modification_found = False
            post_dominators = self.__get_post_dominators(node_list)
            path_splits = self.__get_path_splits(node_list)
            merge_nodes = self.__get_merge_nodes(path_splits, post_dominators)

            for split_node in merge_nodes:
                if merge_nodes[split_node] is None:
                    # no merge exists -> no merge necessary since a return is encountered
                    continue

                # save information for later linearization step
                branch_entry_to_exit: Dict[int, int] = dict()
                branch_entry_to_updated_entry: Dict[int, int] = dict()
                branch_exit_to_updated_exit: Dict[int, int] = dict()
                empty_branches: Set[Tuple[int, int]] = set()

                for branch_entry in get_successors(self.graph, split_node):
                    for branch_exit in get_predecessors(self.graph, cast(int, merge_nodes[split_node])):
                        if branch_exit == split_node and branch_entry == merge_nodes[split_node]:
                            empty_branches.add((branch_entry, branch_exit))

                        if branch_entry not in post_dominators:
                            continue
                        if branch_exit in post_dominators[branch_entry]:
                            branch_entry_to_exit[branch_entry] = branch_exit
                if len(empty_branches) != 0:
                    raise ValueError("No empty branches allowed!")
                # initialize dicts
                for branch_entry in get_successors(self.graph, split_node):
                    branch_entry_to_updated_entry[branch_entry] = branch_entry
                for branch_exit in get_predecessors(self.graph, cast(int, merge_nodes[split_node])):
                    branch_exit_to_updated_exit[branch_exit] = branch_exit

                # create context snapshot
                context_snapshot_id = self.get_new_node_id()
                self.graph.add_node(context_snapshot_id, data=ContextSnapshot(context_snapshot_id, self.experiment))
                node_list.append(context_snapshot_id)
                add_temporary_edge(self.graph, split_node, context_snapshot_id)

                # create branch entries
                for successor in get_successors(self.graph, split_node):
                    # create and connect context restore node
                    branch_context_restore_id = self.get_new_node_id()
                    self.graph.add_node(
                        branch_context_restore_id,
                        data=ContextRestore(branch_context_restore_id, self.experiment),
                    )
                    node_list.append(branch_context_restore_id)
                    add_temporary_edge(self.graph, context_snapshot_id, branch_context_restore_id)

                    # redirect branch to restore node
                    redirect_edge(self.graph, split_node, branch_context_restore_id, successor, successor)
                    branch_entry_to_updated_entry[successor] = branch_context_restore_id
                    modification_found = True

                # create and connect create merge node and snapshot pop node
                context_merge_node_id = self.get_new_node_id()
                self.graph.add_node(context_merge_node_id, data=ContextMerge(context_merge_node_id, self.experiment))
                context_snapshot_pop_node_id = self.get_new_node_id()
                self.graph.add_node(
                    context_snapshot_pop_node_id, data=ContextSnapshotPop(context_snapshot_pop_node_id, self.experiment)
                )
                add_temporary_edge(self.graph, context_merge_node_id, context_snapshot_pop_node_id)
                add_temporary_edge(self.graph, context_snapshot_pop_node_id, cast(int, merge_nodes[split_node]))
                node_list.append(context_merge_node_id)
                node_list.append(context_snapshot_pop_node_id)

                # create branch exits
                for predecessor in get_predecessors(self.graph, cast(int, merge_nodes[split_node])):
                    # create context save node
                    context_save_node_id = self.get_new_node_id()
                    self.graph.add_node(context_save_node_id, data=ContextSave(context_save_node_id, self.experiment))
                    node_list.append(context_save_node_id)
                    # connect context save node
                    add_temporary_edge(self.graph, context_save_node_id, context_merge_node_id)
                    redirect_edge(
                        self.graph, predecessor, predecessor, cast(int, merge_nodes[split_node]), context_save_node_id
                    )
                    branch_exit_to_updated_exit[predecessor] = context_save_node_id

                # linearize the branched section by concatenating all branches
                last_node_id = context_snapshot_id
                for entry in branch_entry_to_exit:
                    if last_node_id != context_snapshot_id:
                        self.graph.remove_edge(last_node_id, context_merge_node_id)
                    exit = branch_entry_to_exit[entry]
                    updated_entry = branch_entry_to_updated_entry[entry]
                    updated_exit = branch_exit_to_updated_exit[exit]
                    # redirect entry
                    redirect_edge(self.graph, context_snapshot_id, last_node_id, updated_entry, updated_entry)
                    last_node_id = updated_exit

                #            # linearize empty branches
                #            for entry,  exit in empty_branches:
                #                updated_entry = branch_entry_to_updated_entry[entry]
                #                updated_exit = branch_exit_to_updated_exit[updated_entry]
                #                print("EMPTY: ", entry, exit)
                #                # redirect entry
                #                redirect_edge(self.graph, context_snapshot_id, last_node_id, updated_entry, updated_entry)
                #                last_node_id = updated_exit

                # connect exit of linearized section
                # add_temporary_edge(self.graph, last_node_id, context_merge_node_id)

                # DEBUG
                if modification_found:
                    break
            convert_temporary_edges(self.graph)

    def __get_merge_nodes(
        self, path_splits: Set[int], post_dominators: Dict[int, Set[int]]
    ) -> Dict[int, Optional[int]]:
        """Calculates and returns the merge nodes for paths starting a the given node"""
        merge_nodes: Dict[int, Optional[int]] = dict()
        for node in path_splits:
            successors = get_successors(self.graph, node)
            candidates = post_dominators[node]
            for succ in successors:
                candidates = candidates.intersection(post_dominators[succ])

            # cleanup candidates to get the earliest merge
            modification_found = True
            while modification_found:
                modification_found = False
                for c1 in candidates:
                    for c2 in candidates:
                        if c1 == c2:
                            continue
                        if c1 in post_dominators[c2]:
                            candidates.remove(c1)
                            modification_found = True
                            break
                    if modification_found:
                        break

            # prepare return value
            if len(candidates) == 0:
                merge_nodes[node] = None
            elif len(candidates) == 1:
                merge_nodes[node] = list(candidates)[0]
            else:
                print("More than one merge node identified for path split: " + str(node) + " : " + str(candidates))
                show(self.graph)
                raise ValueError(
                    "More than one merge node identified for path split: " + str(node) + " : " + str(candidates)
                )
        return merge_nodes

    def __get_path_splits(self, node_list: List[int]) -> Set[int]:
        """Returns nodes which result in path splits"""
        path_splits: Set[int] = set()
        for node in node_list:
            if len(get_successors(self.graph, node)) > 1:
                path_splits.add(node)
        return path_splits

    def __get_post_dominators(self, node_list: List[int]) -> Dict[int, Set[int]]:
        """calculates and returns the post dominators for all nodes in the function"""
        post_dominators: Dict[int, Set[int]] = dict()
        # initialize
        modified: Set[int] = set()
        for node in node_list:
            post_dominators[node] = {node}
            modified.add(node)

        # iterate
        while len(modified) > 0:
            new_modified: Set[int] = set()
            for node in modified:
                predecessors = get_predecessors(self.graph, node)
                for pred in predecessors:
                    len_pre = len(post_dominators[pred])
                    post_dominators[pred] = post_dominators[pred].union(post_dominators[node])
                    if len_pre != len(post_dominators[pred]):
                        new_modified.add(pred)
            modified = new_modified
        return post_dominators

    def __parse_branched_sections(self):
        """Branched sections in the CU Graph are represented by a serialized version in the MOG.
        To make this possible, Context Snapshot, Restore and Merge points are added to allow a synchronization
        'between' the different branches"""
        visited_nodes: Set[int] = set()
        for idx, function_node in enumerate(self.pet.all_nodes(FunctionNode)):
            print("parsing ", function_node.name, idx, "/", len(self.pet.all_nodes(FunctionNode)))
            _, _ = self.__parse_raw_node(self.cu_id_to_graph_node_id[function_node.id], visited_nodes)

        # remove visited nodes, since duplicates exist now
        for node_id in visited_nodes:
            self.graph.remove_node(node_id)

    def __parse_raw_node(self, root_node_id: int, visited_nodes: Set[int]) -> Tuple[int, int]:
        """parses the nodes starting from the root node and returns the id of the first and last node in the created linearized path"""
        visited_nodes.add(root_node_id)
        # if this node has children, parse them before proceeding
        children_entry_points: List[int] = []
        for child_id in get_children(self.graph, root_node_id):
            child_entry_point, _ = self.__parse_raw_node(child_id, visited_nodes)
            children_entry_points.append(child_entry_point)

        if len(get_successors(self.graph, root_node_id)) <= 1:
            return self.__parse_path_node(root_node_id, children_entry_points, visited_nodes)
        else:
            return self.__parse_branching_point(root_node_id, children_entry_points, visited_nodes)

    def __parse_path_node(
        self, root_node_id: int, connect_to_children: List[int], visited_nodes: Set[int]
    ) -> Tuple[int, int]:
        """parses a regular path node, i.e. one with at most one successor node.
        returns the id of the first and last node in the created linearized path"""

        # create a duplicate of the root node
        duplicate_node_id = self.get_new_node_id()
        tmp_node_data = data_at(self.graph, root_node_id)
        self.graph.add_node(duplicate_node_id, data=tmp_node_data)

        # connect duplicated entry node to children
        for children_id in connect_to_children:
            add_child_edge(self.graph, duplicate_node_id, children_id)

        if len(get_successors(self.graph, root_node_id)) == 0:
            # root is a path end
            return duplicate_node_id, duplicate_node_id
        else:
            # proceed to successor
            successor_id = get_successors(self.graph, root_node_id)[0]
            path_entry, path_end = self.__parse_raw_node(successor_id, visited_nodes)

            # connect created duplicate to the successor
            add_temporary_edge(self.graph, duplicate_node_id, path_entry)

            return duplicate_node_id, path_end

    def __parse_branching_point(
        self, root_node_id: int, connect_to_children: List[int], visited_nodes: Set[int]
    ) -> Tuple[int, int]:
        """parses a branching point, i.e. a node with more than one successor.
        returns the id of the first and last node in the created linearized path"""

        #########
        # Step 1: create duplicate of root node
        # Step 2: create and connect context snapshot
        # Step 3: parse branches
        #   Step 3.1: create and connect context restore node
        #   Step 3.2: parse branch
        #   Step 3.3: create context save node
        #   Step 3.3: connect restore and save not to branch entry and exit
        # step 4: create and connect context merge node
        # Step 5: create and connect context snapshot pop
        ########

        # Step 1: create duplicate of root node
        duplicate_node_id = self.get_new_node_id()
        tmp_node_data = data_at(self.graph, root_node_id)
        self.graph.add_node(duplicate_node_id, data=tmp_node_data)

        # Step 2: create and connect context snapshot
        context_snapshot_id = self.get_new_node_id()
        self.graph.add_node(context_snapshot_id, data=ContextSnapshot(context_snapshot_id, self.experiment))
        add_temporary_edge(self.graph, duplicate_node_id, context_snapshot_id)

        # Step 3: parse branches
        last_added_node_id = context_snapshot_id
        for successor in get_successors(self.graph, root_node_id):
            # Step 3.1: create and connect context restore node
            branch_context_restore_id = self.get_new_node_id()
            self.graph.add_node(
                branch_context_restore_id,
                data=ContextRestore(branch_context_restore_id, self.experiment),
            )
            add_temporary_edge(self.graph, last_added_node_id, branch_context_restore_id)

            # Step 3.2: parse branch
            branch_entry, branch_exit = self.__parse_raw_node(successor, visited_nodes)

            # Step 3.3: create context save node
            branch_context_save_id = self.get_new_node_id()
            self.graph.add_node(branch_context_save_id, data=ContextSave(branch_context_save_id, self.experiment))

            # Step 3.3: connect restore and save node to branch entry and exit
            add_temporary_edge(self.graph, branch_context_restore_id, branch_entry)
            add_temporary_edge(self.graph, branch_exit, branch_context_save_id)

            # update last_added_node_id
            last_added_node_id = branch_context_save_id

        # step 4: create and connect context merge node
        context_merge_node_id = self.get_new_node_id()
        self.graph.add_node(context_merge_node_id, data=ContextMerge(context_merge_node_id, self.experiment))
        add_temporary_edge(self.graph, last_added_node_id, context_merge_node_id)

        # Step 5: create and connect context snapshot pop
        context_snapshot_pop_id = self.get_new_node_id()
        self.graph.add_node(
            context_snapshot_pop_id,
            data=ContextSnapshotPop(context_snapshot_pop_id, self.experiment),
        )
        add_temporary_edge(self.graph, context_merge_node_id, context_snapshot_pop_id)

        # connect duplicated entry node to children
        for children_id in connect_to_children:
            add_child_edge(self.graph, duplicate_node_id, children_id)

        return duplicate_node_id, context_snapshot_pop_id

    def __add_cu_nodes(self):
        """adds Workload nodes which represent the CU Nodes to the graph.
        The added nodes will not be connected in any way."""
        for cu_node in self.pet.all_nodes(CUNode):
            # create new node for CU
            new_node_id = self.get_new_node_id()
            self.cu_id_to_graph_node_id[cu_node.id] = new_node_id
            # calculate accessed data
            written_memory_regions, read_memory_regions = get_data_accesses_for_cu(self.pet, cu_node.id)
            try:
                parallelizable_workload = calculate_workload(
                    self.pet, cu_node, ignore_function_calls_and_cached_values=True
                )
            except RecursionError as rerr:
                warnings.warn("Cost calculation not possible for node: " + str(cu_node.id) + "!")
                parallelizable_workload = 0
            self.graph.add_node(
                new_node_id,
                data=Workload(
                    node_id=new_node_id,
                    experiment=self.experiment,
                    cu_id=cu_node.id,
                    sequential_workload=0,
                    parallelizable_workload=parallelizable_workload,
                    written_memory_regions=written_memory_regions,
                    read_memory_regions=read_memory_regions,
                ),
            )

    def __add_loop_nodes(self):
        """adds Loop Nodes to the graph.
        connects contained nodes using Children edges"""
        for loop_node in self.pet.all_nodes(LoopNode):
            # calculate metadata
            loop_subtree = self.pet.subtree_of_type(loop_node, CUNode)

            # get iteration count for loop
            if loop_node.loop_data is not None:
                iteration_count = max(1, loop_node.loop_data.average_iteration_count)
            else:
                # loop has not been executed, thus, no information was gathered during the profiling
                # set iteration_count to prevent divisions by zero
                iteration_count = 1

            # create new node for Loop
            new_node_id = self.get_new_node_id()
            self.cu_id_to_graph_node_id[loop_node.id] = new_node_id
            try:
                discopop_workload = calculate_workload(
                    self.pet, loop_node, ignore_function_calls_and_cached_values=True
                )
            except RecursionError as rerr:
                warnings.warn("Cost calculation not possible for node: " + str(loop_node.id) + "!")
                discopop_workload = 0
            self.graph.add_node(
                new_node_id,
                data=Loop(
                    node_id=new_node_id,
                    cu_id=loop_node.id,
                    discopop_workload=discopop_workload,
                    iterations=iteration_count,
                    position=loop_node.start_position(),
                    experiment=self.experiment,
                ),
            )
            # connect loop node and entry node via a child edge
            entry_node_cu_id = loop_node.get_entry_node(self.pet).id
            add_child_edge(self.graph, new_node_id, self.cu_id_to_graph_node_id[entry_node_cu_id])

            # redirect edges from outside the loop to the entry node to the Loop node
            for s, t, d in self.pet.in_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(s) not in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[s],
                        new_source_id=self.cu_id_to_graph_node_id[s],
                        old_target_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_target_id=new_node_id,
                    )

            # redirect edges to the outside of the loop
            for s, t, d in self.pet.out_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(t) not in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_source_id=new_node_id,
                        old_target_id=self.cu_id_to_graph_node_id[t],
                        new_target_id=self.cu_id_to_graph_node_id[t],
                    )

            # copy entry node
            entry_node_id = self.cu_id_to_graph_node_id[entry_node_cu_id]
            entry_node_data = data_at(self.graph, entry_node_id)
            copied_entry_node_id = self.get_new_node_id()
            self.graph.add_node(copied_entry_node_id, data=entry_node_data)

            # redirect edges from inside the loop to the copy of the entry node
            for s, t, d in self.pet.in_edges(entry_node_cu_id, EdgeType.SUCCESSOR):
                if self.pet.node_at(s) in loop_subtree:
                    redirect_edge(
                        self.graph,
                        old_source_id=self.cu_id_to_graph_node_id[s],
                        new_source_id=self.cu_id_to_graph_node_id[s],
                        old_target_id=self.cu_id_to_graph_node_id[entry_node_cu_id],
                        new_target_id=copied_entry_node_id,
                    )

            # redirect accesses to the cu_id of the entry node to the newly created loop node
            self.cu_id_to_graph_node_id[entry_node_cu_id] = new_node_id

    def __add_functions(self):
        """parse function nodes in the PET graph.
        Results in the creation of a forest of function graphs."""
        for function_node in self.pet.all_nodes(FunctionNode):
            # create function root node and register it in the graph
            new_node_id = self.get_new_node_id()
            self.graph.add_node(
                new_node_id,
                data=FunctionRoot(
                    node_id=new_node_id,
                    experiment=self.experiment,
                    cu_id=function_node.id,
                    name=function_node.name,
                ),
            )
            # connect function node to its entry node
            add_child_edge(
                self.graph,
                new_node_id,
                self.cu_id_to_graph_node_id[function_node.get_entry_cu_id(self.pet)],
            )
            # save ID
            self.cu_id_to_graph_node_id[function_node.id] = new_node_id

    def __add_pet_successor_edges(self):
        for cu_node in self.pet.all_nodes(CUNode):
            for successor_cu_id in [t for s, t, d in self.pet.out_edges(cu_node.id, EdgeType.SUCCESSOR)]:
                add_successor_edge(
                    self.graph,
                    self.cu_id_to_graph_node_id[cu_node.id],
                    self.cu_id_to_graph_node_id[successor_cu_id],
                )

    def __mark_branch_affiliation(self):
        """Mark each nodes' branch affiliation to allow a simple check for 'on same branch' relation
        without considering the successor relation."""

        def mark_branched_section(node, branch_stack):
            node_data = data_at(self.graph, node)
            if isinstance(node_data, ContextSnapshot):
                branch_stack.append(node)
            elif isinstance(node_data, ContextSnapshotPop):
                branch_stack.pop()
            # mark current node
            node_data.branch_affiliation = copy.deepcopy(branch_stack)

            # mark children
            for child in get_children(self.graph, node):
                mark_branched_section(child, copy.deepcopy(branch_stack))

            # mark successors (at most one successor can exist,
            # since parallelization suggestions have not been imported yet)
            for successor in get_successors(self.graph, node):
                mark_branched_section(successor, copy.deepcopy(branch_stack))

        for function_node in get_all_function_nodes(self.graph):
            current_node = get_children(self.graph, function_node)[0]
            mark_branched_section(current_node, [])

    def __calculate_data_flow(self):
        self.in_data_flow = dict()
        self.out_data_flow = dict()

        def inlined_data_flow_calculation(current_node, current_last_writes):
            while current_node is not None:
                # check if current_node uses written data
                reads, writes = get_read_and_written_data_from_subgraph(
                    self.graph, current_node, ignore_successors=True
                )

                for mem_reg in current_last_writes:
                    # check if incoming data flow exists
                    if mem_reg in reads or mem_reg in writes:
                        # uses written data
                        # create incoming data flow
                        if current_node not in self.in_data_flow:
                            self.in_data_flow[current_node] = set()
                        self.in_data_flow[current_node].add(current_last_writes[mem_reg])
                        # create outgoing data flow
                        if current_last_writes[mem_reg] not in self.out_data_flow:
                            self.out_data_flow[current_last_writes[mem_reg]] = set()
                        self.out_data_flow[current_last_writes[mem_reg]].add(current_node)
                # save write to calculate outgoing data flow
                for mem_reg in writes:
                    current_last_writes[mem_reg] = current_node

                # inline children
                for child in get_children(self.graph, current_node):
                    current_last_writes = inlined_data_flow_calculation(child, current_last_writes)

                # continue to successor
                successors = get_successors(self.graph, current_node)
                if len(successors) > 1:
                    raise ValueError("only a single successor should exist at this stage in the process!")

                elif len(successors) == 1:
                    current_node = successors[0]
                else:
                    current_node = None

            return current_last_writes

        # Note: at this point in time, the graph MUST NOT have branched sections
        for function_node in get_all_function_nodes(self.graph):
            last_writes: Dict[MemoryRegion, int] = dict()
            inlined_data_flow_calculation(get_children(self.graph, function_node)[0], last_writes)

        for key in self.out_data_flow:
            for entry in self.out_data_flow[key]:
                if not self.graph.has_edge(key, entry):
                    add_dataflow_edge(self.graph, key, entry)

    def __propagate_reads_and_writes(self):
        # initialize queue
        queue = [n for n in self.graph.nodes]

        while queue:
            current = queue.pop()
            current_data = data_at(self.graph, current)
            parents = get_all_parents(self.graph, current)
            queue += [p for p in parents if p not in queue]
            for p in parents:
                data_at(self.graph, p).written_memory_regions.update(current_data.written_memory_regions)
                data_at(self.graph, p).read_memory_regions.update(current_data.read_memory_regions)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, List, Tuple

import networkx as nx  # type: ignore
from sympy import Integer  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.context.ContextObjectUtils import (
    get_transfer_costs,
)
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot


def add_data_transfer_costs(
    graph: nx.DiGraph,
    function_performance_models: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]],
    environment: Experiment,
) -> Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]]:
    """Calculates the data transfer costs for each of the given performance models and adds them to the respective model."""
    result_dict: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]] = dict()

    for function in function_performance_models:
        result_dict[function] = []
        for cost_model, context in function_performance_models[function]:
            # calculate costs of data transfers
            # For now, it is assumed, that only a single data transfer happens at once
            # and no asynchronous transfers happen.
            # todo: This should be extended in the future.
            data_transfer_costs = get_transfer_costs(context, environment=environment)
            # extend the cost_model
            cost_model = cost_model.parallelizable_plus_combine(data_transfer_costs)

            # add the updated entry to result_dict
            result_dict[function].append((cost_model, context))
    return result_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import os
from typing import Optional
from discopop_explorer.pattern_detectors.PatternBase import PatternBase

from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum

from ..PEGraphX import LoopNode, Node, NodeID, LineID, PEGraphX
from ..utils import calculate_workload, calculate_per_iteration_workload_of_loop


class PatternInfo(PatternBase):
    """Base class for pattern detection info"""

    iterations_count: int
    average_iteration_count: int
    entries: int
    instructions_count: Optional[int]
    workload: Optional[int]
    per_iteration_workload: Optional[int]
    dp_optimizer_device_id: Optional[int] = None  # used by discopop_optimizer. unused by discopop_explorer.
    device_id: Optional[int]
    device_type: Optional[DeviceTypeEnum]

    def __init__(self, node: Node):
        """
        :param node: node, where pipeline was detected
        """
        PatternBase.__init__(self, node)

        self.average_iteration_count = (
            node.loop_data.average_iteration_count
            if (isinstance(node, LoopNode) and node.loop_data is not None)
            else -1
        )
        self.iterations_count = (
            node.loop_data.total_iteration_count if (isinstance(node, LoopNode) and node.loop_data is not None) else -1
        )
        self.entries = node.loop_data.entry_count if (isinstance(node, LoopNode) and node.loop_data is not None) else -1

        # TODO self.instructions_count = total_instructions_count(pet, node)
        self.workload = None
        self.per_iteration_workload: Optional[int] = None
        # TODO self.workload = calculate_workload(pet, node)

        self.device_id = None
        self.device_type = None

    def get_workload(self, pet: PEGraphX) -> int:
        """returns the workload of self._node"""
        if self.workload is not None:
            return self.workload
        try:
            self.workload = calculate_workload(pet, self._node)
        except RecursionError as rerr:
            import warnings

            warnings.warn("Cost calculation not possible for node: " + str(self._node.id) + "!")
            self.workload = 0

        return self.workload

    def get_per_iteration_workload(self, pet: PEGraphX) -> int:
        """returns the per iteration workload of self"""
        if self.per_iteration_workload is not None:
            return self.per_iteration_workload
        self.per_iteration_workload = calculate_per_iteration_workload_of_loop(pet, self._node)
        return self.per_iteration_workload



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import tkinter as tk
from pathlib import Path
from typing import Optional, List, cast

from discopop_library.CodeGenerator.classes.ContentBuffer import ContentBuffer
from discopop_library.CodeGenerator.classes.Line import Line


class CodePreviewLine(Line):
    highlight_color: Optional[str]

    def __init__(self, parent_line_num: int, line_num=None, content=""):
        super().__init__(parent_line_num, line_num, content)
        self.highlight_color = None

    def display(self, wizard, parent_element: tk.Text, line_idx: int, max_line_num: int):
        if not self.content.endswith("\n"):
            self.content += "\n"

        # assemble line_num_str if requested
        line_num_str = ""
        if wizard.settings.code_preview_show_line_numbers == 1:
            line_num_str = str(self.line_num) if self.line_num is not None else ""
            while len(line_num_str) < len(str(max_line_num)):
                line_num_str += " "
            line_num_str += "  "  # padding

        # assemble line for display
        line = ""
        line += line_num_str
        line += self.content

        parent_element.insert(tk.END, line)

        # highlight inserted pragmas
        if self.line_num is None:
            # todo generate background colors
            background_color = "#e5f2b3" if line_idx % 2 == 0 else "#a4ed9a"
            if self.owns_region is not None:
                # highlight entire line if a new region is created
                self.__highlight(parent_element, line_idx, 0, len(line), background_color)
            else:
                # highlight pragma only
                self.__highlight(
                    parent_element,
                    line_idx,
                    len(line) - len(self.content),
                    len(line),
                    background_color,
                )

    def __highlight(
        self,
        parent_element: tk.Text,
        line_idx: int,
        start_position: int,
        end_position: int,
        color: str,
    ):
        """highlights the given section of the line in the given color"""
        start_position_str = str(line_idx) + "." + str(start_position)
        end_position_str = str(line_idx) + "." + str(end_position)
        parent_element.tag_add(
            color + ":" + start_position_str + "-" + end_position_str,
            start_position_str,
            end_position_str,
        )
        parent_element.tag_config(
            color + ":" + start_position_str + "-" + end_position_str,
            background=color,
            foreground="black",
        )


class CodePreviewContentBuffer(ContentBuffer):
    max_line_num: int  # used to determine width of padding
    file_id: int
    next_free_region_id = 0
    lines: List[Line]

    def __init__(self, wizard, file_id: int, source_code_path: Path, tab_width: int = 4):
        super().__init__(file_id, source_code_path, tab_width, line_type=CodePreviewLine)
        self.wizard = wizard

    def show_in(self, parent_element: tk.Text):
        """Displays the contents of the CodePreview object in the supplied parent_element."""
        for line_idx, line in enumerate(self.lines):
            # offset line_id to account for start with 1
            offset_line_id = line_idx + 1
            cast(CodePreviewLine, line).display(self.wizard, parent_element, offset_line_id, self.max_line_num)

    def jump_to_first_modification(self, parent_element: tk.Text):
        """Jumps to the location of the first modified source code location."""
        first_location = ""
        for idx, line in enumerate(self.lines):
            if line.line_num is None:
                first_location = "" + str(idx) + ".0"
                break
        if first_location == "":
            first_location = "1.0"
        parent_element.see(first_location)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import tkinter as tk
from tkinter import ttk


class ScrollableTextWidget(object):
    frame: tk.Frame
    text_container: tk.Text

    def __init__(self, parent_frame):
        self.frame = ttk.Frame(parent_frame)
        self.frame.pack(fill=tk.BOTH, expand=True)
        self.frame.rowconfigure(0, weight=1)
        self.frame.columnconfigure(0, weight=1)

        # create content frame and scroll bars
        self.text_container = tk.Text(self.frame, wrap=tk.NONE)
        self.text_container.grid(
            row=0,
            column=0,
            sticky="nsew",
        )

        # create a Scrollbar and associate it with the content frame
        y_scrollbar = ttk.Scrollbar(self.frame, command=self.text_container.yview)
        y_scrollbar.grid(row=0, column=1, sticky="nsew")
        self.text_container["yscrollcommand"] = y_scrollbar.set
        x_scrollbar = ttk.Scrollbar(self.frame, orient="horizontal", command=self.text_container.xview)
        x_scrollbar.grid(row=1, column=0, columnspan=2, sticky="nsew")
        self.text_container["xscrollcommand"] = x_scrollbar.set
        self.text_container.config(state=tk.DISABLED)

    def set_text(self, content: str):
        self.text_container.config(state=tk.NORMAL)
        self.text_container.delete("1.0", tk.END)
        self.text_container.insert("1.0", content)
        self.text_container.update()
        self.text_container.config(state=tk.DISABLED)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
from typing import Dict, List, Optional, Tuple, cast
import warnings

from sympy import Expr, Integer, Symbol
import sympy
from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.CostModels.DataTransfer.DataTransferCosts import add_data_transfer_costs
from discopop_library.discopop_optimizer.CostModels.utilities import (
    get_node_performance_models,
    get_performance_models_for_functions,
)
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.utilities.MOGUtilities import get_all_function_nodes, show
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


def evaluate_configuration(
    experiment: Experiment,
    decisions: List[int],
    arguments: OptimizerArguments,
) -> Tuple[Tuple[int, ...], Expr, ContextObject]:
    """Evaluate the configuration specified by the decisions for the current set of substitutions.
    Returns the used decisions and the calculated costs as a tuple.
    Note: To compare values across ranges of system specifications, use the ranges obtainable via System.get_symbol_values_and_distributions
    to update the substitutions and execute evaluate_configuration for each set of values."""
    result = Expr(-42)
    # get main function
    main_function: Optional[FunctionRoot] = None
    funciton_node_ids = get_all_function_nodes(experiment.optimization_graph)
    function_root_nodes = [cast(FunctionRoot, data_at(experiment.optimization_graph, fid)) for fid in funciton_node_ids]
    for function_id in funciton_node_ids:
        if cast(FunctionRoot, data_at(experiment.optimization_graph, function_id)).name == "main":
            main_function = cast(FunctionRoot, data_at(experiment.optimization_graph, function_id))
    if main_function is None:
        raise ValueError("No main function found!")

    function_performance_models_without_context = get_performance_models_for_functions(
        experiment, experiment.optimization_graph, restrict_to_decisions=set(decisions)
    )

    function_performance_models = calculate_data_transfers(
        experiment.optimization_graph, function_performance_models_without_context, experiment
    )
    function_performance_models = add_data_transfer_costs(
        experiment.optimization_graph,
        function_performance_models,
        experiment,
    )

    selected_function_models: Dict[FunctionRoot, Tuple[CostModel, ContextObject]] = dict()
    for function in function_performance_models:
        if len(function_performance_models[function]) > 1:
            warnings.warn("Selection for function:" + function.name + " not unambiguous!")
            for m, c in function_performance_models[function]:
                print("--> ", m.path_decisions)
        elif len(function_performance_models[function]) < 1:
            raise ValueError("No generated models for function:", function.name, " for decisions: ", decisions)
        selected_function_models[function] = function_performance_models[function][0]

    # apply selected substitutions
    # collect substitutions
    local_substitutions = copy.deepcopy(experiment.substitutions)
    for function in function_performance_models:
        # register substitution
        local_substitutions[cast(Symbol, function.sequential_costs)] = selected_function_models[function][
            0
        ].sequential_costs
        local_substitutions[cast(Symbol, function.parallelizable_costs)] = selected_function_models[function][
            0
        ].parallelizable_costs

    result_model = copy.deepcopy(selected_function_models[main_function][0])
    result_context = copy.deepcopy(selected_function_models[main_function][1])

    # perform iterative substitutions
    modification_found = True
    while modification_found:
        modification_found = False
        # apply substitution to parallelizable costs
        tmp_model = result_model.parallelizable_costs.subs(local_substitutions)
        if tmp_model != result_model.parallelizable_costs:
            modification_found = True
        result_model.parallelizable_costs = tmp_model

        # apply substitutions to sequential costs
        tmp_model = result_model.sequential_costs.subs(local_substitutions)
        if tmp_model != result_model.sequential_costs:
            modification_found = True
        result_model.sequential_costs = result_model.sequential_costs.subs(local_substitutions)

    # replace Expr(0) with 0
    result_model.sequential_costs = result_model.sequential_costs.subs({Expr(Integer(0)): Integer(0)})
    result_model.parallelizable_costs = result_model.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})

    # unmark replaced free symbols
    local_sorted_free_symbols = copy.deepcopy(experiment.sorted_free_symbols)
    local_free_symbol_ranges = copy.deepcopy(experiment.free_symbol_ranges)
    for symbol in experiment.substitutions:
        if symbol in experiment.free_symbols:
            experiment.free_symbols.remove(symbol)
        if symbol in local_free_symbol_ranges:
            del local_free_symbol_ranges[symbol]
        if symbol in local_sorted_free_symbols:
            local_sorted_free_symbols.remove(symbol)

    # calculate costs
    result = sympy.re(result_model.parallelizable_costs + result_model.sequential_costs) + sympy.im(
        result_model.parallelizable_costs + result_model.sequential_costs
    )

    return (tuple(decisions), result, result_context)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Dict, List, Tuple, Set, Optional

from discopop_explorer.PEGraphX import PEGraphX, EdgeType, NodeID, CUNode, MemoryRegion

global_write_unique_id = 0


def initialize_writes(
    memory_region_liveness: Dict[MemoryRegion, List[NodeID]],
    written_memory_regions_by_cu: Dict[NodeID, Set[MemoryRegion]],
) -> Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]:
    global global_write_unique_id

    result_dict: Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]] = dict()
    for mem_reg in memory_region_liveness:
        if mem_reg not in result_dict:
            result_dict[mem_reg] = set()
        for cu_id in memory_region_liveness[mem_reg]:
            written = False
            if cu_id in written_memory_regions_by_cu:
                if mem_reg in written_memory_regions_by_cu[cu_id]:
                    written = True
            optional_unique_write_id = None
            if written:
                optional_unique_write_id = global_write_unique_id
                global_write_unique_id += 1

            result_dict[mem_reg].add((cu_id, optional_unique_write_id))

    return result_dict


def propagate_writes(
    comb_gpu_reg, pet: PEGraphX, writes: Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]
) -> Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]:
    """propagate writes to parents.
     propagate writes to successors and their children.
     propagate writes to calling functions if a return has been reached.
    Stop on the level of the base function"""
    parent_functions: List[NodeID] = []
    for region in comb_gpu_reg.contained_regions:
        parent_functions.append(pet.get_parent_function(pet.node_at(region.node_id)).id)

    modification_found = True
    while modification_found:
        # propagate writes to parents
        modification_found = True
        while modification_found:
            modification_found = False
            for mem_reg in writes:
                for cu_id_1, write_identifier_1 in writes[mem_reg]:
                    if write_identifier_1 is None:
                        # no write
                        continue
                    # propagate to parents
                    for parent_id, _, _ in pet.in_edges(cu_id_1, EdgeType.CHILD):
                        if parent_id in parent_functions:
                            # do not propagate above the function which contains the GPU Regions
                            continue
                        if (parent_id, write_identifier_1) not in writes[mem_reg]:
                            writes[mem_reg].add((parent_id, write_identifier_1))
                            modification_found = True
                            # propagated to parent
                            break
                    if modification_found:
                        break
                if modification_found:
                    break

        # propagate to successors and their children
        modification_found = True
        while modification_found:
            modification_found = False
            for mem_reg in writes:
                for cu_id_1, write_identifier_1 in writes[mem_reg]:
                    if write_identifier_1 is None:
                        # no write
                        continue
                    # propagate to successors
                    for _, successor_id, _ in pet.out_edges(cu_id_1, EdgeType.SUCCESSOR):
                        if (successor_id, write_identifier_1) not in writes[mem_reg]:
                            writes[mem_reg].add((successor_id, write_identifier_1))
                            modification_found = True
                            # propagated to successor

                            # propagate to children of successors
                            for _, child_id, _ in pet.out_edges(successor_id, EdgeType.CHILD):
                                if (child_id, write_identifier_1) not in writes[mem_reg]:
                                    writes[mem_reg].add((child_id, write_identifier_1))
                                    modification_found = True
                                    # propagated to children of successor
                            # propagate to called functions of successors
                            for _, called_node_id, _ in pet.out_edges(successor_id, EdgeType.CALLSNODE):
                                if (called_node_id, write_identifier_1) not in writes[mem_reg]:
                                    writes[mem_reg].add((called_node_id, write_identifier_1))
                                    modification_found = True
                                    # propagated to called nodes of successor
                            break
                    if modification_found:
                        break
                if modification_found:
                    break

        # propagate writes to calling functions if a return has been reached.
        modification_found = True
        while modification_found:
            modification_found = False
            for mem_reg in writes:
                for cu_id, write_identifier in writes[mem_reg]:
                    pet_node = pet.node_at(cu_id)
                    if type(pet_node) != CUNode:
                        continue
                    if pet_node.return_instructions_count > 0:
                        # propagate write to calling cus
                        parent_function = pet.get_parent_function(pet_node)
                        callees = [s for s, t, d in pet.in_edges(parent_function.id, EdgeType.CALLSNODE)]
                        for callee_id in callees:
                            if (callee_id, write_identifier) not in writes[mem_reg]:
                                writes[mem_reg].add((callee_id, write_identifier))
                                modification_found = True
                                # propagated to callee
                        if modification_found:
                            break
                if modification_found:
                    break

    return writes


def cleanup_writes(
    writes: Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]
) -> Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]:
    """remove entries from the Sets with an second-element entry of None,
    if it is overwritten by a different memory write."""
    for mem_reg in writes:
        to_be_removed: Set[Tuple[NodeID, Optional[int]]] = set()
        # determine elements to be removed
        for cu_id, ident in writes[mem_reg]:
            if ident is None:
                continue
            if (cu_id, None) in writes[mem_reg]:
                # entry is overwritten by write access (cu_id, ident)
                to_be_removed.add((cu_id, None))
        # remove elements
        for cu_id, ident in to_be_removed:
            if (cu_id, ident) in writes[mem_reg]:
                writes[mem_reg].remove((cu_id, ident))
    return writes


def group_writes_by_cu(
    writes: Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]]
) -> Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]:
    result_dict: Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]] = dict()

    for mem_reg in writes:
        for cu_id, ident in writes[mem_reg]:
            if cu_id not in result_dict:
                result_dict[cu_id] = dict()
            if mem_reg not in result_dict[cu_id]:
                result_dict[cu_id][mem_reg] = set()
            result_dict[cu_id][mem_reg].add(ident)
    return result_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from dataclasses import dataclass


@dataclass
class CodeGeneratorArguments(object):
    """Container Class for the arguments passed to the discopop_code_generator"""

    fmap: str
    json: str
    patterns: str
    outputdir: str
    skip_compilation_check: bool
    compile_check_command: str

    def __post_init__(self):
        self.__validate()

    def __validate(self):
        pass

    def __str__(self):
        return str(self.__dict__)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from discopop_library.discopop_optimizer.classes.nodes.GenericNode import GenericNode


import networkx as nx  # type: ignore


from typing import cast


def data_at(graph: nx.DiGraph, node_id: int) -> GenericNode:
    """Return the data object stored at the networkx node with id node_id."""
    return cast(GenericNode, graph.nodes[node_id]["data"])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
import sys
from typing import Dict, Set, Tuple, Optional, List, cast

import networkx as nx  # type: ignore
from networkx import NetworkXNoCycle, MultiDiGraph

from discopop_explorer.PEGraphX import (
    PEGraphX,
    EdgeType,
    NodeType,
    NodeID,
    MemoryRegion,
    CUNode,
    Dependency,
    FunctionNode,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import UpdateType
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Update import Update


class Context(object):
    cu_id: NodeID
    device_id: int
    seen_writes_by_device: Dict[int, Dict[MemoryRegion, Set[Tuple[Optional[int], Optional[NodeID]]]]]

    def __init__(self, cu_id: NodeID, device_id: int):
        self.cu_id = cu_id
        self.device_id = device_id
        self.seen_writes_by_device = dict()
        self.print()

    def __str__(self):
        result_str = ""
        result_str += "Context:\n"
        result_str += "\tcu_id: " + str(self.cu_id) + "\n"
        result_str += "\tdevice: " + str(self.device_id) + "\n"
        result_str += "\twrites:\n"
        for device_id in self.seen_writes_by_device:
            result_str += "\t\tDEVICE: " + str(device_id) + "\n"
            for mem_reg in self.seen_writes_by_device[device_id]:
                result_str += "\t\t\t" + mem_reg + ": "
                for ident in self.seen_writes_by_device[device_id][mem_reg]:
                    result_str += str(ident) + ", "
                result_str += "\n"
            result_str += "\n"
        return result_str

    def print(self):
        print(self, file=sys.stderr)

    def update_writes(
        self,
        device_id_to_update: int,
        writes: Dict[MemoryRegion, Set[Optional[int]]],
        origin_cu_id: NodeID,
    ) -> Set[Tuple[MemoryRegion, bool]]:
        """Returns the updated memory regions"""
        updated_memory_regions: Set[Tuple[MemoryRegion, bool]] = set()

        if device_id_to_update not in self.seen_writes_by_device:
            self.seen_writes_by_device[device_id_to_update] = dict()
        for mem_reg in writes:
            is_initialization = (
                True
                if (mem_reg not in self.seen_writes_by_device[device_id_to_update])
                and (mem_reg not in self.seen_writes_by_device[self.device_id])
                else False
            )
            if mem_reg not in self.seen_writes_by_device[device_id_to_update]:
                self.seen_writes_by_device[device_id_to_update][mem_reg] = set()
                print("==> added ", mem_reg, " to device: ", device_id_to_update, file=sys.stderr)
            for ident in writes[mem_reg]:
                if ident not in [t[0] for t in self.seen_writes_by_device[device_id_to_update][mem_reg]]:
                    # list of seen writes will be updated
                    updated_memory_regions.add((mem_reg, is_initialization))
                    print(
                        "--> adding memreg: ",
                        mem_reg,
                        " ident: ",
                        ident,
                        " initialization? ",
                        is_initialization,
                        file=sys.stderr,
                    )
                print("ADD : ", (ident, origin_cu_id), file=sys.stderr)
                # todo: do not simply add, but replace current entry with origin_cu_id if it is a successor of the current entry
                #  necessary, since relying on set behavior is not sufficient anymore
                if ident not in [t[0] for t in self.seen_writes_by_device[device_id_to_update][mem_reg]]:
                    self.seen_writes_by_device[device_id_to_update][mem_reg].add((ident, origin_cu_id))
        return updated_memory_regions

    def find_required_updates(self, pet: PEGraphX, new_device_id: int) -> Set[Tuple[MemoryRegion, int, int, NodeID]]:
        # update required, if seen writes of new device is not a superset of old device id
        required_updates: Set[Tuple[MemoryRegion, int, int, NodeID]] = set()

        # check if unsynchronized changes exist
        device_id_1 = self.device_id
        device_id_2 = new_device_id

        overlapping_mem_reg = [
            mem_reg
            for mem_reg in self.seen_writes_by_device[device_id_1]
            if mem_reg in self.seen_writes_by_device[device_id_2]
        ]
        for mem_reg in overlapping_mem_reg:
            # issue update, if writes of new device is not a superset of old device
            missing_write_identifiers = [
                (ident, origin)
                for (ident, origin) in self.seen_writes_by_device[device_id_1][mem_reg]
                if ident not in [t[0] for t in self.seen_writes_by_device[device_id_2][mem_reg]] and ident is not None
            ]
            if len(missing_write_identifiers) > 0:
                # get position of the last write
                last_write_location: Optional[NodeID] = None
                for ident, origin in missing_write_identifiers:
                    if last_write_location is None:
                        last_write_location = origin
                    if pet.is_predecessor(cast(NodeID, last_write_location), cast(NodeID, origin)):
                        last_write_location = origin

                required_updates.add((mem_reg, device_id_1, device_id_2, cast(NodeID, last_write_location)))

        return required_updates

    def synchronize_states(self, new_device_id: int):
        overlapping_mem_reg = [
            mem_reg
            for mem_reg in self.seen_writes_by_device[self.device_id]
            if mem_reg in self.seen_writes_by_device[new_device_id]
        ]
        for mem_reg in overlapping_mem_reg:
            # synchronize from old to new device
            # issue update, if writes of new device is not a superset of old device
            for ident, origin in self.seen_writes_by_device[self.device_id][mem_reg]:
                if ident not in [t[0] for t in self.seen_writes_by_device[new_device_id][mem_reg]]:
                    print("SYNCHRONIZED: ", mem_reg, (ident, origin), file=sys.stderr)
                self.seen_writes_by_device[new_device_id][mem_reg].add((ident, origin))

    def request_updates_from_other_devices(self, pet, new_device_id: int) -> Set[Tuple[MemoryRegion, int, int, NodeID]]:
        required_updates: Set[Tuple[MemoryRegion, int, int, NodeID]] = set()
        to_be_removed = set()
        for mem_reg in self.seen_writes_by_device[new_device_id]:
            # check if updates on other devices exist
            for other_device_id in self.seen_writes_by_device:
                if other_device_id == new_device_id:
                    continue
                if mem_reg not in self.seen_writes_by_device[other_device_id]:
                    continue
                # update exist, if entries for new_device_id are not a superset of other_device_id
                missing_identifiers = [
                    (ident, origin)
                    for (ident, origin) in self.seen_writes_by_device[other_device_id][mem_reg]
                    if ident not in [t[0] for t in self.seen_writes_by_device[new_device_id][mem_reg]]
                ]
                if len(missing_identifiers) > 0:
                    # get position of the last write
                    last_write_location: Optional[NodeID] = None
                    for ident, origin in missing_identifiers:
                        if last_write_location is None:
                            last_write_location = origin
                        if pet.is_predecessor(cast(NodeID, last_write_location), cast(NodeID, origin)):
                            last_write_location = origin

                    required_updates.add((mem_reg, other_device_id, new_device_id, cast(NodeID, last_write_location)))

                # remove none identifier
                for ident, origin in self.seen_writes_by_device[new_device_id][mem_reg]:
                    if ident is None:
                        to_be_removed.add((new_device_id, mem_reg, (ident, origin)))

        # update seen writes according to required updates
        for mem_reg, other_device_id, new_device_id, last_write_location in required_updates:
            missing_identifiers = [
                (ident, origin)
                for (ident, origin) in self.seen_writes_by_device[other_device_id][mem_reg]
                if ident not in [t[0] for t in self.seen_writes_by_device[new_device_id][mem_reg]]
            ]

            print("ADDING: ", missing_identifiers, " TO ", mem_reg, file=sys.stderr)

            self.seen_writes_by_device[new_device_id][mem_reg].update(missing_identifiers)

        for entry in to_be_removed:
            if entry[2] in self.seen_writes_by_device[entry[0]][entry[1]]:
                self.seen_writes_by_device[entry[0]][entry[1]].remove(entry[2])
        return required_updates

    def update_to(
        self,
        pet: PEGraphX,
        next_cu_id: NodeID,
        comb_gpu_reg,
        writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]],
    ) -> Set[Update]:
        print("UPDATE TO: ", next_cu_id, file=sys.stderr)

        required_updates: Set[Update] = set()
        # cuid is always the CUID currently saved in the context

        # calculate device id of new cu
        new_device_id = get_device_id(comb_gpu_reg, next_cu_id)

        # check for device switch
        device_switch_occurred = True if new_device_id != self.device_id else False

        # apply updates to written memory regions
        writes_for_update = dict()
        if next_cu_id in writes_by_device[new_device_id]:
            writes_for_update = writes_by_device[new_device_id][next_cu_id]
        self_updated_memory_regions = self.update_writes(new_device_id, writes_for_update, next_cu_id)
        print("UPDATED MEMORY REGIONS: ", self_updated_memory_regions, file=sys.stderr)

        # identify required updates
        updated_memory_regions = self.find_required_updates(pet, new_device_id)

        # synchronize states between old and new device
        self.synchronize_states(new_device_id)

        # check if update has to be requested from other device due to a read (identifier: None)
        requested_updates = self.request_updates_from_other_devices(pet, new_device_id)

        # report data movement
        for mem_reg, from_device, to_device, last_write_to_mem_reg in updated_memory_regions:
            update_direction = get_update_type(from_device, to_device)
            print(
                "REPORTED! ",
                update_direction,
                " -> ",
                str(mem_reg),
                from_device,
                "->",
                to_device,
                " last_write@",
                last_write_to_mem_reg,
                file=sys.stderr,
            )
            required_updates.add(
                Update(
                    self.cu_id,
                    next_cu_id,
                    {mem_reg},
                    update_direction,
                    {mem_reg: last_write_to_mem_reg},
                )
            )

        # report data movement
        for mem_reg, from_device, to_device, last_write_to_mem_reg in requested_updates:
            update_direction = get_update_type(from_device, to_device)
            print(
                "REQUESTED! ",
                update_direction,
                " -> ",
                str(mem_reg),
                from_device,
                "->",
                to_device,
                " last_write@",
                last_write_to_mem_reg,
                file=sys.stderr,
            )
            required_updates.add(
                Update(
                    self.cu_id,
                    next_cu_id,
                    {mem_reg},
                    update_direction,
                    {mem_reg: last_write_to_mem_reg},
                )
            )

        # update the context
        self.cu_id = next_cu_id
        self.device_id = new_device_id

        print("DONE: \n", file=sys.stderr)

        return required_updates


def get_device_id(comb_gpu_reg, cu_id: NodeID) -> int:
    if cu_id in comb_gpu_reg.device_cu_ids:
        return 1
    return 0


def __identify_merge_node(pet, successors: List[NodeID]) -> Optional[NodeID]:
    def check_validity_of_potential_merge_node(node_id: NodeID):
        # return True if the given node is a valid merge node.
        # return False otherwise.
        # do not allow return BB's as merge nodes, since this would be trivially true for every path split
        potential_merge_node = pet.node_at(node_id)
        if type(potential_merge_node) != CUNode:
            return False
        if (
            "return" in str(potential_merge_node.basic_block_id)
            and potential_merge_node.end_position() == pet.get_parent_function(potential_merge_node).end_position()
        ):
            # do not consider return BB as merge node
            return False
        return True

    if len(successors) == 0:
        raise ValueError("Empty list of successors!")

    parent_function = pet.get_parent_function(pet.node_at(successors[0]))
    post_dominators = parent_function.get_immediate_post_dominators(pet)

    # initialize lists of current post dominators
    current_post_dominators: List[NodeID] = [post_dominators[node_id] for node_id in successors]
    # initialize lists of visited post dominators for each node in "successors"
    visited_post_dominators: List[Set[NodeID]] = []
    for idx, cpd in enumerate(current_post_dominators):
        # add the node itself as well as it's first post dominator to visited nodes for the initialization
        visited_post_dominators.append({cpd, successors[idx]})

    if len(visited_post_dominators) == 0 and len(current_post_dominators) == 0:
        raise ValueError("No post dominators found!")

    # check for common post dominators
    while True:
        # check if a common post dominator exists
        common_post_dominators = visited_post_dominators[0]
        for idx, vpd_set in enumerate(visited_post_dominators):
            common_post_dominators.intersection_update(vpd_set)
            if len(common_post_dominators) == 0:
                break

        if len(common_post_dominators) > 0:
            # return identified merge node
            for potential_merge_node_id in common_post_dominators:
                if check_validity_of_potential_merge_node(potential_merge_node_id):
                    return potential_merge_node_id

        if len(current_post_dominators) == 0:
            # No merge node found and no node to be checked anymore. Exit the loop.
            break

        # if not, add current post dominators to the visited list and resolve each post dominator by one step
        # break if new post dominator is equal to the old (end of path reached)
        for idx, cpd in enumerate(current_post_dominators):
            visited_post_dominators[idx].add(cpd)
        new_post_dominators = []
        for cpd in current_post_dominators:
            if cpd not in post_dominators:
                post_dominators[cpd] = cpd
            tmp = post_dominators[cpd]
            if tmp == cpd:
                # end of path reached, do not add tmp to list of new post dominators
                continue
            # get the next post dominator
            new_post_dominators.append(tmp)
        # replace the list of current post dominators (~ frontier)
        current_post_dominators = new_post_dominators
    # no merge node found
    return None


def identify_updates(
    comb_gpu_reg,
    pet: PEGraphX,
    writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]],
    unrolled_function_graph: MultiDiGraph,
) -> Set[Update]:
    identified_updates: Set[Update] = set()
    # get parent functions
    parent_functions: Set[NodeID] = set()
    for region in comb_gpu_reg.contained_regions:
        parent_functions.add(pet.get_parent_function(pet.node_at(region.node_id)).id)

    for parent_function_id in parent_functions:
        print("IDENTIFY UPDATES FOR: ", pet.node_at(parent_function_id).name, file=sys.stderr)
        # determine entry points
        entry_points: List[NodeID] = []
        for function_child_id in [t for s, t, d in pet.out_edges(parent_function_id, EdgeType.CHILD)]:
            in_successor_edges = pet.in_edges(function_child_id, EdgeType.SUCCESSOR)
            if len(in_successor_edges) == 0 and pet.node_at(function_child_id).type == NodeType.CU:
                entry_points.append(function_child_id)

        for entry_point in entry_points:
            print(
                "\tEntry point: ",
                entry_point,
                "@L.",
                pet.node_at(entry_point).start_line,
                file=sys.stderr,
            )

            # create a new context
            # todo add contexts support for multiple devices
            device_id = get_device_id(comb_gpu_reg, entry_point)
            context = Context(entry_point, device_id)

            try:
                entry_point_writes = writes_by_device[device_id][entry_point]
            except KeyError:
                entry_point_writes = dict()

            context.update_writes(device_id, entry_point_writes, entry_point)

            # start calculation of updates for entry_point
            identified_updates.update(
                __calculate_updates(
                    pet,
                    comb_gpu_reg,
                    context,
                    entry_point,
                    writes_by_device,
                    set(),
                    unrolled_function_graph,
                )
            )

    return identified_updates


def get_update_type(from_device_id: int, to_device_id: int) -> UpdateType:
    if from_device_id == 0 and to_device_id != 0:
        return UpdateType.TO_DEVICE
    elif from_device_id != 0 and to_device_id == 0:
        return UpdateType.FROM_DEVICE
    raise ValueError(
        "Unsupported update type requested for device IDS: " + str(from_device_id) + " -> " + str(to_device_id)
    )


def __calculate_updates(
    pet: PEGraphX,
    comb_gpu_reg,
    context: Context,
    cur_node_id: NodeID,
    writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]],
    visited_nodes: Set[NodeID],
    unrolled_function_graph: MultiDiGraph,
) -> Set[Update]:
    # calculate and return updates between ctx.cu_id and its immediate successor cur_node
    identified_updates: Set[Update] = set()

    # pass on calculation
    end_reached = False
    while not end_reached:
        context.print()

        # todo: update context to current node and save required updates
        required_updates = context.update_to(pet, cur_node_id, comb_gpu_reg, writes_by_device)
        identified_updates.update(required_updates)

        # calculate successors of current node
        # successors = [cast(NodeID, t) for s, t, d in pet.out_edges(cur_node_id, EdgeType.SUCCESSOR)]
        successors = [t for s, t, d in unrolled_function_graph.out_edges(cur_node_id, data="data")]

        if len(successors) == 0:
            end_reached = True
            continue
        elif len(successors) == 1:
            if cur_node_id == successors[0]:
                # cycle! break
                end_reached = True
                continue
            # start calculation for successor of current node
            cur_node_id = successors[0]
            print("PROCEED TO ", cur_node_id, file=sys.stderr)
            continue
        else:
            # multiple successors exist

            # determine merge node if existing
            if False:
                # Disabled merge node calculation, since new graph structures are free of cycles!
                # Keep the old code for now.
                print("SUCCESSORS: ", successors)
                merge_node = __identify_merge_node(pet, successors)
                print("MERGE NODE: ", merge_node)

                # if merge node exists, skip branched section and treat it as a single node
                if merge_node is not None:
                    cur_node_id = merge_node
                    print("PROCEED TO MERGE NODE ", cur_node_id, file=sys.stderr)
                    continue
            else:
                merge_node = None

            # if no merge node exists, start recursion calculation of updates for each branch
            if merge_node is None:
                # recursive calculation will stop at the end of the funciton body, thus return identified updates
                # without executing a new loop iteration
                for successor in successors:
                    if successor in visited_nodes:
                        # prevent endless cycles
                        continue
                    # enter recursion
                    print("ENTER RECURSION FOR: ", successor, file=sys.stderr)
                    tmp_visited_set = copy.deepcopy(visited_nodes)
                    tmp_visited_set.add(successor)

                    identified_updates.update(
                        __calculate_updates(
                            pet,
                            comb_gpu_reg,
                            copy.deepcopy(context),
                            successor,
                            writes_by_device,
                            tmp_visited_set,
                            unrolled_function_graph,
                        )
                    )
                return identified_updates

    return identified_updates


def create_circle_free_function_graphs(pet: PEGraphX, add_dummy_node=True):
    """Remove loops from the CUGraph by unrolling loops in the successor graphs of each function."""
    import networkx as nx

    unrolled_function_graphs: Dict[FunctionNode, nx.MultiDiGraph] = dict()

    # initialize the function graphs
    for function in pet.all_nodes(type=FunctionNode):
        # get subgraph for function
        function_subgraph = pet.g.subgraph(function.children_cu_ids).copy()
        # remove all but successor edges
        to_be_removed = set()
        for edge in function_subgraph.edges:
            edge_data = cast(Dependency, function_subgraph.edges[edge]["data"])
            if edge_data.etype != EdgeType.SUCCESSOR:
                to_be_removed.add(edge)
        for edge in to_be_removed:
            function_subgraph.remove_edge(edge[0], edge[1])
        # store function subgraph
        unrolled_function_graphs[function] = function_subgraph

    # UNROLLING ALGORITHM:
    # for function:
    # while cycle in function_graph:
    # unroll found cycle
    # ==> Identify cycle entries
    # ==> Identify cycle exits
    # create branch equivalent to "not entering the cycle"
    # create branch equivalent to "entering the cycle"
    # TODO insert a second iteration of the cycle?

    # unroll each function separately
    for function in unrolled_function_graphs:
        print("UNROLLING FUNCTION: ", function.name)
        try:
            cycle_edges = nx.find_cycle(unrolled_function_graphs[function])
            print("\tCycle: ", cycle_edges)
        except NetworkXNoCycle:
            print("\tNo cycle found.")
            continue
        # while cycle in graph:
        exit_node_buffer = []
        while len(cycle_edges) > 0:
            cycle_nodes: Set[NodeID] = set()
            for s, t, d in cycle_edges:
                cycle_nodes.add(s)
                cycle_nodes.add(t)
            # unroll cycle
            # ==> Identify cycle entries
            entry_nodes: Set[NodeID] = set()
            for node_id, _, _ in cycle_edges:
                predecessors = [s for s, t, d in pet.in_edges(node_id, EdgeType.SUCCESSOR)]
                if len(predecessors) == 0:
                    entry_nodes.add(node_id)
                    continue
                predecessors = [s for s in predecessors if s not in cycle_nodes]
                if len(predecessors) > 0:
                    entry_nodes.add(node_id)

            # ==> Identify cycle exits
            for potential_exit_node in set([s for s, t, d in cycle_edges] + [t for s, t, d in cycle_edges]):
                print("cyc: ", cycle_nodes)
                print("exit: ", potential_exit_node)
                potential_cycle_successor_nodes = [
                    t for s, t, d in pet.out_edges(potential_exit_node, EdgeType.SUCCESSOR) if t not in cycle_nodes
                ]
                print("POT: ", potential_cycle_successor_nodes)

                # only consider such cycle_successors which DO NOT share a direct parent with the potential_exit_node, except for functions
                pen_parents = [
                    s
                    for s, t, d in pet.in_edges(potential_exit_node, EdgeType.CHILD)
                    if type(pet.node_at(s)) != FunctionNode
                ]
                filtered_pcsn = []
                for pcsn in potential_cycle_successor_nodes:
                    pcsn_parents = [
                        s for s, t, d in pet.in_edges(pcsn, EdgeType.CHILD) if type(pet.node_at(s)) != FunctionNode
                    ]
                    if len([nid for nid in pen_parents if nid in pcsn_parents]) == 0:
                        # no shared parents
                        filtered_pcsn.append(pcsn)
                    # second chance: add pcsn, if it has a successor outside the given parent
                    for successor in pet.direct_successors(pet.node_at(pcsn)):
                        valid = True
                        for parent in [s for s, t, d in pet.in_edges(successor.id)]:
                            if parent in pcsn_parents:
                                valid = False
                                break
                        if valid:
                            filtered_pcsn.append(pcsn)
                            break

                # prevent endless cycling
                if potential_exit_node in exit_node_buffer:
                    filtered_pcsn = potential_cycle_successor_nodes

                if len(filtered_pcsn) == 0:
                    # not a valid exit node, add to buffer and skip it
                    exit_node_buffer.append(potential_exit_node)
                    continue

                # reset buffer, since action will be performed
                exit_node_buffer = []

                print("EXITS: ", filtered_pcsn)

                # create branch equivalent to "not entering the cycle"
                # --> nothing to be done, already represented by the edge "potential_exit_node --> cycle_successor"

                # create branch equivalent to "entering the cycle"
                # --> create a copy of the exit node

                if add_dummy_node:
                    unrolled_function_graphs[function].add_node(
                        "dummy:" + potential_exit_node,
                        data=unrolled_function_graphs[function].nodes[potential_exit_node]["data"],
                    )

                # --> redirect the last edge in the cycle to cycle_successor, and insert the created copy into the path
                buffer = set()
                for cycle_edge in cycle_edges:
                    if cycle_edge[1] == potential_exit_node:
                        print("Redirecting: ", cycle_edge, " TO (", cycle_edge[0], ", ", end="")
                        unrolled_function_graphs[function].remove_edge(cycle_edge[0], cycle_edge[1], cycle_edge[2])
                        if add_dummy_node:
                            unrolled_function_graphs[function].add_edge(
                                cycle_edge[0],
                                "dummy:" + potential_exit_node,
                                type=EdgeType.SUCCESSOR,
                            )
                        else:
                            buffer.add(cycle_edge[0])

                for cycle_successor in filtered_pcsn:
                    if add_dummy_node:
                        print("dummy:" + potential_exit_node, ",", end="")
                    print(
                        cycle_successor,
                        ")",
                    )
                    if add_dummy_node:
                        unrolled_function_graphs[function].add_edge("dummy:" + potential_exit_node, cycle_successor)
                    else:
                        for buffered_node_id in buffer:
                            unrolled_function_graphs[function].add_edge(
                                buffered_node_id, cycle_successor, type=EdgeType.SUCCESSOR
                            )

            # prepare next iteration
            try:
                cycle_edges = nx.find_cycle(unrolled_function_graphs[function])
                print("\tCycle: ", cycle_edges)
            except NetworkXNoCycle:
                print("\tNo cycle found.")
                # break the unrolling loop
                cycle_edges = []

    return unrolled_function_graphs


def add_accesses_from_called_functions(
    pet: PEGraphX,
    writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]],
    force_called_functions_to_host: bool = False,
) -> Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]]:
    """Gather written and read memory regions on a function level.
    Add the gathered information to calling CU nodes."""
    values_propagated = True
    cycles = 0
    while values_propagated:
        cycles += 1
        values_propagated = False
        for function in pet.all_nodes(type=FunctionNode):
            print("FUNCTION: ", function.name)
            memory_accesses = function.get_memory_accesses(writes_by_device)
            if force_called_functions_to_host:
                for device_id in memory_accesses:
                    if device_id != 0:
                        for mem_reg in memory_accesses[device_id]:
                            if mem_reg not in memory_accesses[0]:
                                memory_accesses[0][mem_reg] = set()
                            memory_accesses[0][mem_reg].update(memory_accesses[device_id][mem_reg])
                to_be_removed = [key for key in memory_accesses if key != 0]
                for key in to_be_removed:
                    del memory_accesses[key]

            # add memory_accesses to calling CU's in writes_by_device
            called_by = [s for s, t, d in pet.in_edges(function.id, EdgeType.CALLSNODE)]
            print("\tcalled by: ", called_by)
            for device_id in memory_accesses:
                if device_id not in writes_by_device:
                    writes_by_device[device_id] = dict()
                for calling_cu_id in called_by:
                    if calling_cu_id not in writes_by_device[device_id]:
                        writes_by_device[device_id][calling_cu_id] = dict()
                    for mem_reg in memory_accesses[device_id]:
                        if mem_reg not in writes_by_device[device_id][calling_cu_id]:
                            writes_by_device[device_id][calling_cu_id][mem_reg] = set()
                        len_pre = len(writes_by_device[device_id][calling_cu_id][mem_reg])
                        writes_by_device[device_id][calling_cu_id][mem_reg].update(memory_accesses[device_id][mem_reg])
                        len_post = len(writes_by_device[device_id][calling_cu_id][mem_reg])
                        values_propagated = (
                            (values_propagated or True) if len_pre < len_post else (values_propagated or False)
                        )
    print("cycles: ", cycles)
    return writes_by_device


def identify_updates_in_unrolled_function_graphs(
    comb_gpu_reg,
    pet: PEGraphX,
    writes_by_device: Dict[int, Dict[NodeID, Dict[MemoryRegion, Set[Optional[int]]]]],
    unrolled_function_graphs: Dict[FunctionNode, MultiDiGraph],
) -> Set[Update]:
    identified_updates: Set[Update] = set()
    # get parent functions
    parent_functions: Set[NodeID] = set()
    for region in comb_gpu_reg.contained_regions:
        parent_functions.add(pet.get_parent_function(pet.node_at(region.node_id)).id)

    for parent_function_id in parent_functions:
        print("IDENTIFY UPDATES FOR: ", pet.node_at(parent_function_id).name, file=sys.stderr)
        # determine entry points
        entry_points: List[NodeID] = []
        for function_child_id in [t for s, t, d in pet.out_edges(parent_function_id, EdgeType.CHILD)]:
            in_successor_edges = pet.in_edges(function_child_id, EdgeType.SUCCESSOR)
            if len(in_successor_edges) == 0 and pet.node_at(function_child_id).type == NodeType.CU:
                entry_points.append(function_child_id)

        for entry_point in entry_points:
            print(
                "\tEntry point: ",
                entry_point,
                "@L.",
                pet.node_at(entry_point).start_line,
                file=sys.stderr,
            )

            # create a new context
            # todo add contexts support for multiple devices
            device_id = get_device_id(comb_gpu_reg, entry_point)
            context = Context(entry_point, device_id)

            try:
                entry_point_writes = writes_by_device[device_id][entry_point]
            except KeyError:
                entry_point_writes = dict()

            context.update_writes(device_id, entry_point_writes, entry_point)

            # start calculation of updates for entry_point
            identified_updates.update(
                __calculate_updates(
                    pet,
                    comb_gpu_reg,
                    context,
                    entry_point,
                    writes_by_device,
                    set(),
                    unrolled_function_graphs[cast(FunctionNode, pet.node_at(parent_function_id))],
                )
            )

    return identified_updates



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os
from typing import Any, Dict, List, Union

from sympy import Integer

from discopop_library.discopop_optimizer.classes.system.devices.CPU import CPU
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.system.devices.GPU import GPU


def generate_default_system_configuration(file_path: str):
    """Generates a system configuration file using the default values if none exists so far."""
    if os.path.exists(file_path):
        return
    # build default system configuration
    system_configuration: Dict[str, Any] = dict()
    devices: List[Dict[str, Union[float, int, Dict[str, float]]]] = []
    # configure host device
    host_device: Dict[str, Union[float, int, Dict[str, float]]] = {
        "device_id": 0,
        "device_type": DeviceTypeEnum.CPU,
        "frequency": 3000000000,
        "processors": 16,
        "threads": 16,
    }
    # configure gpu_1
    gpu_1: Dict[str, Union[float, int, Dict[str, float]]] = {
        "compute_init_delays": {"target_teams_distribute_parallel_for": 0.01},
        "device_id": 1,
        "device_type": DeviceTypeEnum.GPU,
        "frequency": 256000000,
        "processors": 128,
        "teams": 3200,
        "threads": 3200,
        "transfer_init_delays[us]": {
            "target_data_update": 350,
            "target_enter_data": 900,
            "target_exit_data": 40,
            "average": 730,
        },
        "transfer_speeds": {"D2H_MB/s": 1800, "H2D_MB/s": 3600},  # MB/s
    }
    # configure gpu_2
    gpu_2: Dict[str, Union[float, int, Dict[str, float]]] = {
        "compute_init_delays": {"target_teams_distribute_parallel_for": 0.005},
        "device_id": 2,
        "device_type": DeviceTypeEnum.GPU,
        "frequency": 256000000,
        "processors": 128,
        "teams": 3200,
        "threads": 3200,
        "transfer_init_delays[us]": {
            "target_data_update": 710,
            "target_enter_data": 900,
            "target_exit_data": 50,
            "average": 810,
        },
        "transfer_speeds": {"D2H_MB/s": 1900, "H2D_MB/s": 4200},  # MB/s
    }
    # assemble system_configuration
    devices = [host_device, gpu_1, gpu_2]
    system_configuration["devices"] = devices
    system_configuration["host_device"] = host_device["device_id"]

    # output to file
    with open(file_path, "w") as f:
        json.dump(system_configuration, f)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
import warnings

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode


class ContextRestore(ContextNode):
    def __init__(self, node_id: int, experiment):
        super().__init__(node_id, experiment)

    def get_plot_label(self) -> str:
        return str(self.node_id) + "\nCTX\nrestore"

    def get_modified_context(
        self, node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject
    ) -> ContextObject:
        #        # save save_stack in buffer
        #        buffer_save_stack = copy.deepcopy(context.save_stack)
        #        # save snapshot_stack in buffer
        #        buffer_snapshot_stack = copy.deepcopy(context.snapshot_stack)
        #        # restore the latest entry in snapshot_stack
        #        if len(context.snapshot_stack) < 1:
        #            warnings.warn("Context can not be restored before creating a snapshot!")
        #            return context
        #        restored_context: ContextObject = copy.deepcopy(context.snapshot_stack[-1])
        ##
        #        # add buffer to save_stack
        #        restored_context.save_stack += buffer_save_stack
        #        # set snapshot_stack to the buffered contents
        #        restored_context.snapshot_stack = buffer_snapshot_stack
        #        return restored_context

        return context



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from argparse import ArgumentParser
from discopop_library.ConfigProvider.ConfigProviderArguments import ConfigProviderArguments
from discopop_library.ConfigProvider.config_provider import run


def parse_args() -> ConfigProviderArguments:
    """Parse the arguments passed to the discopop_config_provider"""
    parser = ArgumentParser(description="DiscoPoP Config Provider")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments for experimental features. Experimental arguments may or may not be removed or changed in the future.",
    )

    # fmt: off
    parser.add_argument("-b", "--dp-build-dir", action="store_true",
                        help="Return the path to the DiscoPoP build directory")
    parser.add_argument("-s", "--dp-source-dir", action="store_true",
                        help="Return the path to the DiscoPoP source directory")
    parser.add_argument("--llvm-bin-dir", action="store_true",
                        help="Return the path to the LLVM bin directory")
    # EXPERIMENTAL FLAGS:
    # fmt: on

    arguments = parser.parse_args()

    return ConfigProviderArguments(
        return_dp_build_dir=arguments.dp_build_dir,
        return_dp_source_dir=arguments.dp_source_dir,
        return_llvm_bin_dir=arguments.llvm_bin_dir,
    )


def main() -> str:
    arguments = parse_args()
    retval = run(arguments)
    return retval


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
import os
from typing import List

from lxml import objectify, etree  # type: ignore

from discopop_explorer.PEGraphX import LoopNode, PEGraphX
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import line_contained_in_region


def cu_xml_preprocessing(cu_xml: str) -> str:
    """Execute CU XML Preprocessing.
    Returns file name of modified cu xml file.
    :param cu_xml: path to the xml file
    :return: file name of modified cu xml file.
    """
    xml_fd = open(cu_xml)
    xml_content = ""
    for line in xml_fd.readlines():
        if not (line.rstrip().endswith("</Nodes>") or line.rstrip().endswith("<Nodes>")):
            xml_content = xml_content + line

    xml_content = "<Nodes>{0}</Nodes>".format(xml_content)

    parsed_cu = objectify.fromstring(xml_content)

    iterate_over_cus = True  # used to enable re-starting
    self_added_node_ids: List[str] = []
    while iterate_over_cus:
        used_node_ids = []
        for node in parsed_cu.Node:
            used_node_ids.append(node.get("id"))

        for node in parsed_cu.Node:
            inner_iteration = True
            remaining_recursive_call_in_parent = False
            while inner_iteration:
                used_node_ids = list(set(used_node_ids + self_added_node_ids))

                if node.get("type") == "0":  # iterate over CU nodes
                    # find CU nodes with > 1 recursiveFunctionCalls in own code region
                    if (
                        __preprocessor_cu_contains_at_least_two_recursive_calls(node)
                        or remaining_recursive_call_in_parent
                    ):
                        remaining_recursive_call_in_parent = False
                        # Preprocessor Step 1
                        tmp_cn_entry = None  # (recursiveFunctionCall, nodeCalled)
                        for cne_idx, calls_node_entry in enumerate(node.callsNode):
                            # get first matching entry of node.callsNode
                            try:
                                for rc_idx, rec_call in enumerate(calls_node_entry.recursiveFunctionCall):
                                    rec_call_line = calls_node_entry.nodeCalled[rc_idx].get("atLine")
                                    if str(rec_call_line) in str(rec_call):
                                        tmp_cn_entry = (
                                            rec_call,
                                            calls_node_entry.nodeCalled[rc_idx],
                                        )
                                        break
                            except AttributeError:
                                continue
                        if tmp_cn_entry is None:
                            raise Exception("no matching entries for callsNode found!")

                        parent = node
                        tmp_cn_entry[0].getparent().remove(tmp_cn_entry[0])
                        tmp_cn_entry[1].getparent().remove(tmp_cn_entry[1])
                        parent_copy = copy.copy(parent)
                        parsed_cu.insert(parsed_cu.index(parent), parent_copy)

                        # Preprocessor Step 2 - generate cu id for new element
                        __generate_new_cu_id(parent, parent_copy, used_node_ids, self_added_node_ids)

                        # Preprocessor Step 3
                        parent_copy.callsNode.clear()
                        parent_copy.callsNode.append(tmp_cn_entry[1])
                        parent_copy.callsNode.append(tmp_cn_entry[0])

                        parent_copy.successors.clear()
                        etree.SubElement(parent_copy.successors, "CU")
                        parent_copy.successors.CU._setText(parent.get("id"))

                        # delete childrenNodes-entry from parent
                        tmp_cu_id = tmp_cn_entry[1].text
                        parent.childrenNodes._setText(parent.childrenNodes.text.replace(tmp_cu_id + ",", ""))
                        parent.childrenNodes._setText(parent.childrenNodes.text.replace(tmp_cu_id, ""))

                        # set parent_copy.childrenNodes
                        __set_parent_copy_childrennodes(parent_copy)

                        # Preprocessor Step 4
                        # __remove_overlapping_start_and_end_lines(parent_copy, parent.instructionLines)
                        # __remove_overlapping_start_and_end_lines(parent_copy, parent.readPhaseLines)
                        # __remove_overlapping_start_and_end_lines(parent_copy, parent.writePhaseLines)

                        separator_line = parent.get("startsAtLine")
                        # select smallest recursive function call line >= separator_line + 1
                        parent_new_start_line = None
                        potential_lines = []
                        for tmp1 in parent.callsNode:
                            try:
                                for tmp2 in tmp1.nodeCalled:
                                    try:
                                        potential_lines.append(tmp2.get("atLine"))
                                        pass
                                    except AttributeError:
                                        pass
                            except AttributeError:
                                pass
                        for tmp in potential_lines:
                            if tmp == "":
                                continue
                            if int(tmp[tmp.find(":") + 1 :]) >= int(separator_line[separator_line.find(":") + 1 :]) + 1:
                                if parent_new_start_line is None:
                                    parent_new_start_line = tmp
                                    continue
                                # select smallest instruction line
                                if int(tmp[tmp.find(":") + 1 :]) < int(
                                    parent_new_start_line[parent_new_start_line.find(":") + 1 :]
                                ):
                                    parent_new_start_line = tmp
                        if not potential_lines or (potential_lines and not parent_new_start_line):
                            parent_new_start_line = str(separator_line[: separator_line.index(":")])
                            parent_new_start_line += ":"
                            parent_new_start_line += str(int(separator_line[separator_line.index(":") + 1 :]) + 1)

                        parent.set("startsAtLine", parent_new_start_line)
                        parent_copy.set("endsAtLine", separator_line)

                        # update instruction/readPhase/writePhase lines
                        # __filter_rwi_lines(parent_copy, parent_copy.instructionLines)
                        # __filter_rwi_lines(parent_copy, parent_copy.readPhaseLines)
                        # __filter_rwi_lines(parent_copy, parent_copy.writePhaseLines)

                        # insert separator line to parent_copys instruction,
                        # read and writePhaseLines if not already present
                        # __insert_separator_line(parent_copy, parent_copy.instructionLines)
                        # __insert_separator_line(parent_copy, parent_copy.readPhaseLines)
                        # __insert_separator_line(parent_copy, parent_copy.writePhaseLines)

                        # insert all lines contained in parent to instruction, read and writePhaseLines
                        # __insert_missing_rwi_lines(parent, parent.instructionLines)
                        # __insert_missing_rwi_lines(parent, parent.readPhaseLines)
                        # __insert_missing_rwi_lines(parent, parent.writePhaseLines)

                        # remove returnInstructions if they are not part of the cus anymore
                        __remove_unnecessary_return_instructions(parent_copy)
                        __remove_unnecessary_return_instructions(parent)

                        # add parent.id to parent_function.childrenNodes
                        __add_parent_id_to_children(parsed_cu, parent)

                        # Preprocessor Step 5 (looping)
                        parent_further_cn_entry = None
                        for cne_idx, calls_node_entry in enumerate(parent.callsNode):
                            # get first matching entry of node.callsNode
                            try:
                                for rc_idx, rec_call in enumerate(calls_node_entry.recursiveFunctionCall):
                                    rec_call_line = calls_node_entry.nodeCalled[rc_idx].get("atLine")
                                    if str(rec_call_line) in str(rec_call):
                                        parent_further_cn_entry = (
                                            rec_call,
                                            calls_node_entry.nodeCalled[rc_idx],
                                        )
                                        break
                            except AttributeError:
                                continue
                        if parent_further_cn_entry is None:
                            # parent has no further recursive call, restart outer loop
                            inner_iteration = False
                            continue
                        else:
                            # parent still has recursive calls
                            inner_iteration = True
                            node = parent
                            remaining_recursive_call_in_parent = True
                            continue
                    else:
                        inner_iteration = False
                        continue
                else:
                    # node not of type CU, go to next node
                    inner_iteration = False
                    continue

        iterate_over_cus = False  # disable restarting, preprocessing finished

    # print modified Data.xml to file
    modified_cu_xml = cu_xml.replace(".xml", "-preprocessed.xml")
    if os.path.exists(modified_cu_xml):
        os.remove(modified_cu_xml)
    f = open(modified_cu_xml, "w+")
    f.write(etree.tostring(parsed_cu, pretty_print=True).decode("utf-8"))
    f.close()
    return modified_cu_xml


def __generate_new_cu_id(parent, parent_copy, used_node_ids, self_added_node_ids):
    """Generate the next free CU id and assign it to the parent CU.
    :param parent: parent CU, id will be updated
    :param parent_copy: copy of parent CU (newly created CU)
    :param used_node_ids: list of used cu node id's
    :param self_added_node_ids: list of added node id's"""
    # get next free id for specific tmp_file_id
    parent_copy_id = parent_copy.get("id")
    tmp_file_id = parent_copy_id[: parent_copy_id.index(":")]
    tmp_used_ids = [int(s[s.index(":") + 1 :]) for s in used_node_ids if s.startswith(tmp_file_id + ":")]
    next_free_id = max(tmp_used_ids) + 1
    incremented_id = tmp_file_id + ":" + str(next_free_id)
    parent.set("id", incremented_id)
    self_added_node_ids.append(incremented_id)


def __set_parent_copy_childrennodes(parent_copy):
    """Adds cu nodes called by parent_copy to the childrenNodes list of parent_copy, if not already contained.
    :param parent_copy: cu node to be updated"""
    parent_copy.childrenNodes._setText("")
    for cne_idx, calls_node_entry in enumerate(parent_copy.callsNode):
        try:
            for node_call in calls_node_entry.nodeCalled:
                try:
                    if node_call.text not in parent_copy.childrenNodes.text:
                        parent_copy.childrenNodes._setText(parent_copy.childrenNodes.text + "," + node_call.text)
                        if parent_copy.childrenNodes.text.startswith(","):
                            parent_copy.childrenNodes._setText(parent_copy.childrenNodes.text[1:])
                        if parent_copy.childrenNodes.text.endswith(","):
                            parent_copy.childrenNodes._setText(parent_copy.childrenNodes.text[:-1])
                        continue
                except AttributeError as e1:
                    print(e1)
                    continue
        except AttributeError as e2:
            print(e2)
            continue


def __remove_overlapping_start_and_end_lines(parent_copy, target_list):
    """Removes the first line of parent_copy from parent´s readPhaseLines, writePhaseLines or instructionLines.
    As a result, start and end Lines of both nodes do not overlap anymore.
    :param parent_copy: copy of parent node (newly added node)
    :param target_list: eiter readPhaseLines, writePhaseLines or instructionLines of parent
    """
    try:
        if parent_copy.callsNode.nodeCalled.get("atLine") in target_list.text:
            target_list._setText(target_list.text.replace(parent_copy.callsNode.nodeCalled.get("atLine") + ",", ""))
            target_list._setText(target_list.text.replace(parent_copy.callsNode.nodeCalled.get("atLine"), ""))
            target_list.set("count", str(int(target_list.get("count")) - 1))
    except TypeError:
        target_list._setText(parent_copy.callsNode.nodeCalled.get("atLine"))
        target_list.set("count", "1")


def __filter_rwi_lines(parent_copy, target_list):
    """Removes entries from instructionLines, readPhaseLines and writePhraseLines of parent_copy if their value is not
    between parent_copy.startsAtLine and parent_copy.endsAtLine.
    :param parent_copy: cu node to be filtered
    :param target_list: eiter readPhaseLines, writePhaseLines or instructionLines of parent_copy"""
    try:
        for tmp_line in target_list.text.split(","):
            if not line_contained_in_region(tmp_line, parent_copy.get("startsAtLine"), parent_copy.get("endsAtLine")):
                target_list._setText(target_list.text.replace(tmp_line + ",", ""))
                target_list._setText(target_list.text.replace(tmp_line, ""))
                if target_list.text.endswith(","):
                    target_list._setText(target_list.text[:-1])
                target_list.set("count", str(int(target_list.get("count")) - 1))
    except AttributeError:
        pass


def __insert_separator_line(parent_copy, target_list):
    """Insert separator line to parent_copys instruction, read and writePhaseLines if not already present
    :param parent_copy: cu node to be updated
    :param target_list: eiter readPhaseLines, writePhaseLines or instructionLines of parent_copy"""
    try:
        if not parent_copy.get("endsAtLine") in target_list.text:
            target_list._setText(target_list.text + "," + parent_copy.get("endsAtLine"))
            if target_list.text.startswith(","):
                target_list._setText(target_list.text[1:])
            target_list.set("count", str(int(target_list.get("count")) + 1))
    except TypeError:
        target_list._setText(parent_copy.get("endsAtLine"))
        target_list.set("count", "1")
    target_list._setText(target_list.text.replace(",,", ","))


def __insert_missing_rwi_lines(parent, target_list):
    """Insert all lines contained in parent to instruction, read and writePhaseLines
    :param parent: cu node to be updated
    :param target_list: eiter readPhaseLines, writePhaseLines or instructionLines of parent"""
    cur_line = parent.get("startsAtLine")
    while line_contained_in_region(cur_line, parent.get("startsAtLine"), parent.get("endsAtLine")):
        if cur_line not in target_list.text:
            target_list._setText(cur_line + "," + target_list.text)
            if target_list.text.endswith(","):
                target_list._setText(target_list.text[:-1])
            target_list.set("count", str(int(target_list.get("count")) + 1))
        # increment cur_line by one
        cur_line = cur_line[0 : cur_line.rfind(":") + 1] + str(int(cur_line[cur_line.rfind(":") + 1 :]) + 1)
        continue
    target_list._setText(target_list.text.replace(",,", ","))


def __remove_unnecessary_return_instructions(target):
    """Remove returnInstructions if they are not part of target cu anymore.
    :param target: cu to be checked"""
    if int(target.returnInstructions.get("count")) != 0:
        entries = target.returnInstructions.text.split(",")
        new_entries = []
        for entry in entries:
            if line_contained_in_region(entry, target.get("startsAtLine"), target.get("endsAtLine")):
                new_entries.append(entry)
        target.returnInstructions._setText(",".join(new_entries))
        target.returnInstructions.set("count", str(len(new_entries)))


def __add_parent_id_to_children(parsed_cu, parent):
    """ "Add parent.id to parent_function.childrenNodes
    :param: parsed_cu: parsed contents of cu_xml file
    :param parent: cu node to be added to parent_function's children
    """
    parent_function = None
    for tmp_node in parsed_cu.Node:
        if tmp_node.get("type") == "1":
            if line_contained_in_region(
                parent.get("startsAtLine"), tmp_node.get("startsAtLine"), tmp_node.get("endsAtLine")
            ):
                if line_contained_in_region(
                    parent.get("endsAtLine"),
                    tmp_node.get("startsAtLine"),
                    tmp_node.get("endsAtLine"),
                ):
                    parent_function = tmp_node
                    break
    if parent_function is None:
        print("No parent function found for cu node: ", parent.get("id"), ". Ignoring.")
    else:
        parent_function.childrenNodes._setText(parent_function.childrenNodes.text + "," + parent.get("id"))
        if parent_function.childrenNodes.text.startswith(","):
            parent_function.childrenNodes._setText(parent_function.childrenNodes.text[1:])


def __preprocessor_cu_contains_at_least_two_recursive_calls(node) -> bool:
    """Check if >= 2 recursive function calls are contained in a cu's code region.
    Returns True, if so.
    Returns False, else.
    :param node: CUNode
    :return: bool
    """
    starts_at_line = node.get("startsAtLine").split(":")
    ends_at_line = node.get("endsAtLine").split(":")
    file_id = starts_at_line[0]
    if file_id != ends_at_line[0]:
        raise Exception("error in Data.xml: FileIds of startsAtLine and endsAtLine not matching!")
    starts_at_line = starts_at_line[1]
    ends_at_line = ends_at_line[1]

    # count contained recursive Function calls
    contained_recursive_calls = 0
    for calls_node_entry in node.callsNode:
        try:
            for i in calls_node_entry.recursiveFunctionCall:
                rec_func_calls = [s for s in str(i).split(",") if len(s) > 0]
                if len(rec_func_calls) != 0:
                    for rec_func_call in rec_func_calls:
                        rec_func_call = rec_func_call.split(" ")[1]
                        rfc_file_id = rec_func_call.split(":")[0]
                        rfc_line = rec_func_call.split(":")[1]
                        # test if recursiveFunctionCall is inside CU region
                        if rfc_file_id == file_id and starts_at_line <= rfc_line <= ends_at_line:
                            contained_recursive_calls += 1
        except AttributeError:
            pass
    if contained_recursive_calls >= 2:
        return True
    return False


def check_loop_scopes(pet: PEGraphX):
    """Checks if the scope of loop CUs matches these of their children. Corrects the scope of the loop CU
    (expand only) if necessary
    :param pet: PET graph"""
    for loop_cu in pet.all_nodes(LoopNode):
        for child in pet.direct_children_or_called_nodes(loop_cu):
            if not line_contained_in_region(child.start_position(), loop_cu.start_position(), loop_cu.end_position()):
                # expand loop_cu start_position upwards
                if child.start_line < loop_cu.start_line and loop_cu.file_id == child.file_id:
                    loop_cu.start_line = child.start_line
            if not line_contained_in_region(child.end_position(), loop_cu.start_position(), loop_cu.end_position()):
                # expand loop_cu end_position downwards
                if child.end_line > loop_cu.end_line and loop_cu.file_id == child.file_id:
                    loop_cu.end_line = child.end_line



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import List, Dict, Tuple, Optional, cast

import matplotlib  # type: ignore
import numpy as np
import sympy
from matplotlib import pyplot as plt  # type: ignore
from spb import plot3d, MB, plot  # type: ignore
from sympy import Integer
from sympy import Symbol, Expr

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment


def plot_CostModels(
    experiment,
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
    super_title: Optional[str] = None,
):
    local_sorted_free_symbols = copy.deepcopy(sorted_free_symbols)
    local_free_symbol_ranges = copy.deepcopy(free_symbol_ranges)
    for symbol in experiment.substitutions:
        if symbol in experiment.free_symbols:
            experiment.free_symbols.remove(symbol)
        if symbol in local_free_symbol_ranges:
            del local_free_symbol_ranges[symbol]
        if symbol in local_sorted_free_symbols:
            local_sorted_free_symbols.remove(symbol)

    # apply selected substitutions
    # collect substitutions
    local_substitutions = copy.deepcopy(experiment.substitutions)
    for function in experiment.selected_paths_per_function:
        # register substitution
        local_substitutions[cast(Symbol, function.sequential_costs)] = experiment.selected_paths_per_function[function][
            0
        ].sequential_costs
        local_substitutions[cast(Symbol, function.parallelizable_costs)] = experiment.selected_paths_per_function[
            function
        ][0].parallelizable_costs

    local_models = copy.deepcopy(models)

    # perform iterative substitutions
    modification_found = True
    while modification_found:
        modification_found = False
        for model in local_models:
            # apply substitution to parallelizable costs
            tmp_model = model.parallelizable_costs.subs(local_substitutions)
            if tmp_model != model.parallelizable_costs:
                modification_found = True
            model.parallelizable_costs = tmp_model

            # apply substitutions to sequential costs
            tmp_model = model.sequential_costs.subs(local_substitutions)
            if tmp_model != model.sequential_costs:
                modification_found = True
            model.sequential_costs = model.sequential_costs.subs(local_substitutions)

    # replace Expr(0) with 0
    for model in local_models:
        model.sequential_costs = model.sequential_costs.subs({Expr(Integer(0)): Integer(0)})
        model.parallelizable_costs = model.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})

    if len(local_sorted_free_symbols) == 2:
        __3d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=str(title) + str(super_title) if super_title is not None else title,
        )
    elif len(local_sorted_free_symbols) == 1:
        __2d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=str(title) + str(super_title) if super_title is not None else title,
        )
    elif len(local_sorted_free_symbols) == 0:
        __1d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=title,
            super_title=super_title,
        )
    else:
        print("Plotiting not supported for", len(sorted_free_symbols), "free symbols!")


def plot_CostModels_using_function_path_selections(
    experiment: Experiment,
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
    super_title: Optional[str] = None,
):
    # apply selected substitutions
    # collect substitutions
    local_substitutions = copy.deepcopy(experiment.substitutions)
    for function in experiment.selected_paths_per_function:
        # register substitution
        local_substitutions[cast(Symbol, function.sequential_costs)] = experiment.selected_paths_per_function[function][
            0
        ].sequential_costs
        local_substitutions[cast(Symbol, function.parallelizable_costs)] = experiment.selected_paths_per_function[
            function
        ][0].parallelizable_costs

    local_models = copy.deepcopy(models)

    # perform iterative substitutions
    modification_found = True
    while modification_found:
        modification_found = False
        for model in local_models:
            # apply substitution to parallelizable costs
            tmp_model = model.parallelizable_costs.subs(local_substitutions)
            if tmp_model != model.parallelizable_costs:
                modification_found = True
            model.parallelizable_costs = tmp_model

            # apply substitutions to sequential costs
            tmp_model = model.sequential_costs.subs(local_substitutions)
            if tmp_model != model.sequential_costs:
                modification_found = True
            model.sequential_costs = model.sequential_costs.subs(local_substitutions)

    # replace Expr(0) with 0
    for model in local_models:
        model.sequential_costs = model.sequential_costs.subs({Expr(Integer(0)): Integer(0)})
        model.parallelizable_costs = model.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})

    local_sorted_free_symbols = copy.deepcopy(sorted_free_symbols)
    local_free_symbol_ranges = copy.deepcopy(free_symbol_ranges)
    for symbol in experiment.substitutions:
        if symbol in experiment.free_symbols:
            experiment.free_symbols.remove(symbol)
        if symbol in local_free_symbol_ranges:
            del local_free_symbol_ranges[symbol]
        if symbol in local_sorted_free_symbols:
            local_sorted_free_symbols.remove(symbol)

    if len(local_sorted_free_symbols) == 2:
        __3d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=str(title) + str(super_title) if super_title is not None else title,
        )
    elif len(local_sorted_free_symbols) == 1:
        __2d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=str(title) + str(super_title) if super_title is not None else title,
        )
    elif len(local_sorted_free_symbols) == 0:
        __1d_plot(
            local_models,
            local_sorted_free_symbols,
            local_free_symbol_ranges,
            labels=labels,
            title=title,
            super_title=super_title,
        )
    else:
        print("Plotiting not supported for", len(local_sorted_free_symbols), "free symbols!")


def print_current_function_path_selections(experiment: Experiment):
    """Prints an overview of the currently selected paths for each function to the console"""
    print("###")
    print("SELECTIONS:")
    for function in experiment.selected_paths_per_function:
        print("\t", function.name)
        print("\t\t", experiment.selected_paths_per_function[function][0].path_decisions)
    print("###")


def print_current_substitutions(experiment: Experiment):
    """Prints an overview of the currently selected paths for each function to the console"""
    print("###")
    print("SUBSTITUTIONS:")
    for var in experiment.substitutions:
        print("->", var)
        print("\t", experiment.substitutions[var])
    print("###")


def print_simplified_function(
    experiment: Experiment,
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
    super_title: Optional[str] = None,
):
    """Prints an simplified mathematical function based on the current set of selections"""

    # todo: NOTE: copied from plot_CostModels_using_function_path_selections

    # apply selected substitutions
    # collect substitutions
    local_substitutions = copy.deepcopy(experiment.substitutions)
    for function in experiment.selected_paths_per_function:
        # register substitution
        local_substitutions[cast(Symbol, function.sequential_costs)] = experiment.selected_paths_per_function[function][
            0
        ].sequential_costs
        local_substitutions[cast(Symbol, function.parallelizable_costs)] = experiment.selected_paths_per_function[
            function
        ][0].parallelizable_costs

    local_models = copy.deepcopy(models)

    # perform iterative substitutions
    modification_found = True
    while modification_found:
        modification_found = False
        for model in local_models:
            # apply substitution to parallelizable costs
            tmp_model = model.parallelizable_costs.subs(local_substitutions)
            if tmp_model != model.parallelizable_costs:
                modification_found = True
            model.parallelizable_costs = tmp_model

            # apply substitutions to sequential costs
            tmp_model = model.sequential_costs.subs(local_substitutions)
            if tmp_model != model.sequential_costs:
                modification_found = True
            model.sequential_costs = model.sequential_costs.subs(local_substitutions)

    # replace Expr(0) with 0
    for model in local_models:
        model.sequential_costs = model.sequential_costs.subs({Expr(Integer(0)): Integer(0)})
        model.parallelizable_costs = model.parallelizable_costs.subs({Expr(Integer(0)): Integer(0)})

    local_sorted_free_symbols = copy.deepcopy(sorted_free_symbols)
    local_free_symbol_ranges = copy.deepcopy(free_symbol_ranges)
    for symbol in experiment.substitutions:
        if symbol in experiment.free_symbols:
            experiment.free_symbols.remove(symbol)
        if symbol in local_free_symbol_ranges:
            del local_free_symbol_ranges[symbol]
        if symbol in local_sorted_free_symbols:
            local_sorted_free_symbols.remove(symbol)

    print("###")
    print("FUNCTION:")
    for model in local_models:
        model_costs = sympy.simplify(
            sympy.re(model.parallelizable_costs + model.sequential_costs)
            + sympy.im(model.parallelizable_costs + model.sequential_costs)
        )
        print("-> ", model_costs)
    print("###")


__unique_plot_id = 0


def __1d_plot(
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
    super_title: Optional[str] = None,
):
    global __unique_plot_id
    matplotlib.use("TkAgg")
    # Make a dataset from models:
    height: List[float] = []
    bars: List[str] = []

    for idx, model in enumerate(models):
        model_label = str(model.path_decisions) if labels is None else labels[idx]
        bars.append(model_label)

        # get numeric value from model
        num_value = float(
            sympy.re(model.sequential_costs.evalf() + model.parallelizable_costs.evalf())
            + sympy.im(model.sequential_costs.evalf() + model.parallelizable_costs.evalf())
        )
        height.append(num_value)

    bars_tuple = tuple(bars)
    y_pos = np.arange(len(bars_tuple))

    # Create bars
    plt.figure(__unique_plot_id)
    __unique_plot_id += 1
    if title is not None:
        plt.title(title)  # type: ignore
    if super_title is not None:
        plt.suptitle(super_title)  # type: ignore
    plt.bar(y_pos, height)  # type: ignore
    # Create names on the x-axis
    plt.xticks(y_pos, bars_tuple)
    # Show graphic
    plt.show()


def __2d_plot(
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
):
    matplotlib.use("TkAgg")
    combined_plot = None
    for idx, model in enumerate(models):
        model_label = str(model.path_decisions) if labels is None else labels[idx]
        model_costs = sympy.re(model.parallelizable_costs + model.sequential_costs) + sympy.im(
            model.parallelizable_costs + model.sequential_costs
        )
        if combined_plot is None:
            combined_plot = plot(
                model_costs,
                (
                    sorted_free_symbols[0],
                    free_symbol_ranges[sorted_free_symbols[0]][0],
                    free_symbol_ranges[sorted_free_symbols[0]][1],
                ),
                show=False,
                backend=MB,
                label=model_label,
                ylabel="Execution time [us]",
                title=title,
            )
        else:
            combined_plot.extend(
                plot(
                    model_costs,
                    (
                        sorted_free_symbols[0],
                        free_symbol_ranges[sorted_free_symbols[0]][0],
                        free_symbol_ranges[sorted_free_symbols[0]][1],
                    ),
                    show=False,
                    backend=MB,
                    label=model_label,
                    ylabel="Execution time [us]",
                )
            )
    combined_plot.show()  # type: ignore


def __3d_plot(
    models: List[CostModel],
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    labels: Optional[List[str]] = None,
    title: Optional[str] = None,
):
    matplotlib.use("TkAgg")
    combined_plot = None
    for idx, model in enumerate(models):
        model_label = str(model.path_decisions) if labels is None else labels[idx]
        model_costs = sympy.re(model.parallelizable_costs + model.sequential_costs) + sympy.im(
            model.parallelizable_costs + model.sequential_costs
        )
        if combined_plot is None:
            combined_plot = plot3d(
                model_costs,
                (
                    sorted_free_symbols[0],
                    free_symbol_ranges[sorted_free_symbols[0]][0],
                    free_symbol_ranges[sorted_free_symbols[0]][1],
                ),
                (
                    sorted_free_symbols[1],
                    free_symbol_ranges[sorted_free_symbols[1]][0],
                    free_symbol_ranges[sorted_free_symbols[1]][1],
                ),
                show=False,
                backend=MB,
                label=model_label,
                zlabel="Execution time [us]",
                title=title,
            )
        else:
            combined_plot.extend(
                plot3d(
                    model_costs,
                    (
                        sorted_free_symbols[0],
                        free_symbol_ranges[sorted_free_symbols[0]][0],
                        free_symbol_ranges[sorted_free_symbols[0]][1],
                    ),
                    (
                        sorted_free_symbols[1],
                        free_symbol_ranges[sorted_free_symbols[1]][0],
                        free_symbol_ranges[sorted_free_symbols[1]][1],
                    ),
                    show=False,
                    backend=MB,
                    label=model_label,
                    zlabel="Execution time [us]",
                )
            )
    combined_plot.show()  # type: ignore



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy
from typing import List, Dict, Tuple

import networkx as nx  # type: ignore
from spb import plot3d, MB, plot  # type: ignore
from sympy import Symbol

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.CostModels.utilities import get_random_path
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot


def find_quasi_optimal_using_random_samples(
    experiment: Experiment,
    graph: nx.DiGraph,
    function_root: FunctionRoot,
    random_path_count: int,
    sorted_free_symbols: List[Symbol],
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]],
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution],
    verbose: bool = False,
):
    """Returns the identified minimum, maximum, median, 25% quartile and 75% quartile of the random_path_count samples.
    NOTE: The decisions should be treated as suggestions, not mathematically accurate decisions
    due to the used comparison method!"""
    random_paths: List[Tuple[CostModel, ContextObject]] = []
    if verbose:
        print("Generating ", random_path_count, "random paths")
    i = 0
    # create a temporary copy of the substitutions list to roll back unwanted modifications
    substitutions_buffer = copy.deepcopy(experiment.substitutions)
    while i < random_path_count:
        # reset substitutions
        experiment.substitutions = copy.deepcopy(substitutions_buffer)

        tmp_dict = dict()
        tmp_dict[function_root] = [get_random_path(experiment, graph, function_root.node_id, must_contain=None)]
        try:
            random_paths.append(calculate_data_transfers(graph, tmp_dict, experiment)[function_root][0])
            i += 1
        except ValueError as ve:
            if verbose:
                print(ve)
            # might occur as a result of invalid paths due to restrictions
            pass
    # reset substitutions
    experiment.substitutions = copy.deepcopy(substitutions_buffer)

    # apply substitutions and set free symbol ranges and distributions
    if verbose:
        print("\tApplying substitutions...")

    random_paths_with_substitutions: List[Tuple[CostModel, ContextObject, CostModel]] = []
    for model, context in random_paths:
        substituted_model = copy.deepcopy(model)

        # apply substitutions iteratively
        modification_found = True
        while modification_found:
            modification_found = False
            # apply substitutions to parallelizable costs
            tmp_model = substituted_model.parallelizable_costs.subs(experiment.substitutions)
            if tmp_model != substituted_model.parallelizable_costs:
                modification_found = True
            substituted_model.parallelizable_costs = tmp_model

            # apply substitutions to sequential costs
            tmp_model = substituted_model.sequential_costs.subs(experiment.substitutions)
            if tmp_model != substituted_model.sequential_costs:
                modification_found = True
            substituted_model.sequential_costs = tmp_model
        substituted_model.free_symbol_ranges = free_symbol_ranges
        substituted_model.free_symbol_distributions = free_symbol_distributions

        random_paths_with_substitutions.append((model, context, substituted_model))

    if verbose:
        print("\tSorting...")
    sorted_list = sorted(random_paths_with_substitutions, key=lambda x: x[2])  # BOTTLENECK!
    if verbose:
        print("\tDone.")
    minimum = (sorted_list[0][0], sorted_list[0][1])
    maximum = (sorted_list[-1][0], sorted_list[-1][1])
    median = (sorted_list[int(len(sorted_list) / 2)][0], sorted_list[int(len(sorted_list) / 2)][1])
    upper_quartile = (
        sorted_list[int(len(sorted_list) / 4 * 3)][0],
        sorted_list[int(len(sorted_list) / 4 * 3)][1],
    )
    lower_quartile = (
        sorted_list[int(len(sorted_list) / 4 * 1)][0],
        sorted_list[int(len(sorted_list) / 4 * 1)][1],
    )

    return minimum, maximum, median, lower_quartile, upper_quartile
